#!/usr/bin/env python

# [1] https://pubs.acs.org/doi/full/10.1021/acs.jctc.7b00788
#     libreta: Computerized Optimization and Code Synthesis for Electron
#     Repulsion Integral Evaluation
#     Jun Zhang, 2018
# [2] https://doi.org/10.1002/qua.560400604
#     Two-electron repulsion integrals over Gaussian s functions
#     Gill, Johnson, Pople, 1991

import argparse
import sys

import numpy as np
from scipy.special import factorial2

from pysisyphus.config import CONFIG_DIR, L_MAX, L_AUX_MAX


_BOYS_TABLE = np.load(
    CONFIG_DIR / "wavefunction/ints/boys_table_N_64_xasym_27.1_step_0.01.npy"
)
_BOYS_N_MAX = (
    max(L_MAX, L_AUX_MAX) * 3
)  # Currently, there are at most 3center integrals
_BOYS_X_ASYM = (
    27.0  # Beyond this x the Boys function can be calculated via an analytical formula
)
_BOYS_ADD_POINTS = 10  # Go a bit beyond X_ASYM
_BOYS_STEP = 0.01
_BOYS_FACTOR = 100  # int(1 / _BOYS_FACTOR)
_BOYS_XS = np.arange(0, _BOYS_X_ASYM + (_BOYS_ADD_POINTS + 1) * _BOYS_STEP, _BOYS_STEP)

PI = np.pi
FACT2 = np.array(
    [factorial2(N, exact=True) for N in range(2 * _BOYS_N_MAX)], dtype=np.int64
)


def make_boys_table():
    from scipy.integrate import quad

    def boys(t, x, N):
        return t ** (2 * N) * np.exp(-x * t**2)

    xs = _BOYS_XS
    boys_table = list()
    for N in range(_BOYS_N_MAX + 1):
        print(f"{N=}, calculating at {xs.size} points from {xs[0]} to {xs[-1]}")
        vals = list()
        max_abserr = 0.0
        for i, x in enumerate(xs):
            if i % 1000 == 0:
                print(f"\t{i} values")
            y, abserr = quad(boys, 0, 1, args=(x, N), epsabs=5e-13, epsrel=5e-13)
            max_abserr = max(abserr, max_abserr)
            vals.append(y)
        print(f"\tN={N}, max(abserr)={max_abserr:.8e}")
        boys_table.append(vals)
    boys_table = np.array(boys_table)
    fn = f"boys_table_N_{_BOYS_N_MAX}_xasym_{xs[-1]:.1f}_step_{_BOYS_STEP:.2f}.npy"
    np.save(fn, boys_table)
    return fn


def neville(x, xs_table, table, step, points=5, factor=None):
    """Slow, recursive implementation of Neville interpolation.

    We multiply 'x' by int(1/step), so we only have to deal with integer arithmetic, to
    determine the first entry from the table. 'x_closest' is always chosen in a way, that
    'x' is contained in the 'xs' interval. Just to be sure, one can also suppy the 'factor'
    to this method."""

    if factor is None:
        factor = int(1 / step)

    # Determine closest x
    x_closest = int(x * factor)
    indices = np.arange(points) + x_closest
    xs = xs_table[indices]
    ys = table[indices]
    p_ = {(i, i): ys[i] for i in range(points)}

    def p(i, j):
        key = (i, j)
        if key not in p_:
            xi, xj = xs[[i, j]]
            p_[key] = ((x - xj) * p(i, j - 1) - (x - xi) * p(i + 1, j)) / (xi - xj)
        return p_[key]

    return p(0, points - 1)


def neville_gen5(x, xs_table, table, step, factor=None):
    """Code-generated Neville interpolation using 5 points."""

    points = 5
    if factor is None:
        factor = int(1 / step)

    # Determine closest x
    x_closest = int(x * factor)
    indices = np.arange(points) + x_closest
    xs = xs_table[indices]
    ys = table[indices]

    x0 = -xs[4]
    x1 = x - xs[0]
    x2 = -xs[1]
    x3 = x + x2
    x4 = -xs[2]
    x5 = x + x4
    x6 = -xs[3]
    x7 = x + x6
    x8 = x + x0
    x9 = (x5 * ys[3] - x7 * ys[2]) / (-x6 - xs[2])
    x10 = (x3 * ys[2] - x5 * ys[1]) / (-x4 - xs[1])
    x11 = (-x10 * x7 + x3 * x9) / (-x6 - xs[1])

    return (
        x1
        * (
            -x11 * x8
            + x3
            * (x5 * (x7 * ys[4] - x8 * ys[3]) / (-x0 - xs[3]) - x8 * x9)
            / (-x0 - xs[2])
        )
        / (-x0 - xs[1])
        - x8
        * (
            x1 * x11
            - x7
            * (x1 * x10 - x5 * (x1 * ys[1] - x3 * ys[0]) / (-x2 - xs[0]))
            / (-x4 - xs[0])
        )
        / (-x6 - xs[0])
    ) / (-x0 - xs[0])


def neville_boys(N, x):
    """Wrapper for Boys function from Neville-interpolation."""
    return neville_gen5(
        x, _BOYS_XS, _BOYS_TABLE[N], step=_BOYS_STEP, factor=_BOYS_FACTOR
    )


def factorial_boys(N, x):
    """Boys function for (big) x > 27.0 as described in the SI of [1]. See also [2].
                  ___________
     -N - 1      ╱  -2⋅N - 1
    2      ⋅√π⋅╲╱  x         ⋅(2⋅N - 1)!

    """
    return FACT2[max(0, 2 * N - 1)] / 2 ** (N + 1) * np.sqrt(PI / x ** (2 * N + 1))


def boys(N, xs):
    """Wrapper for Boys function calculation.

    Supports scalar values and np.ndarray for 'xs'.
    """

    def func(N, x):
        return neville_boys(N, x) if (x <= _BOYS_X_ASYM) else factorial_boys(N, x)

    if isinstance(xs, np.ndarray):
        boys_list = list()

        with np.nditer(xs) as it:
            for x in it:
                boys_list.append(func(N, x))
        boys_table = np.reshape(boys_list, xs.shape)
        return boys_table
    else:
        return func(N, xs)


def parse_args(args):
    parser = argparse.ArgumentParser()
    parser.add_argument("--table", action="store_true")
    return parser.parse_args(args)


def run():
    args = parse_args(sys.argv[1:])

    if args.table:
        fn = make_boys_table()
        print(f"Wrote Boys table to '{fn}'.")


if __name__ == "__main__":
    run()



import numpy as np

from pysisyphus.constants import FORCE2ACC
from pysisyphus.dynamics.helpers import energy_forces_getter_closure, remove_com_velocity


def rattle_closure(geom, constraints, dt, tol=1e-3, max_cycles=25,
                   energy_forces_getter=None, remove_com_v=True):
    # Inverse atom masses
    masses = geom.masses
    inv_masses = 1 / masses
    inv_masses_rep = np.repeat(inv_masses, 3)

    dt2 = 0.5 * dt
    dt2_sq = dt2 * dt
    tol_sq = tol**2

    constraints = np.array(constraints, dtype=int)
    inv_masses_sums = inv_masses[constraints[:, 0]] + inv_masses[constraints[:, 1]]
    coords3d = geom.coords3d

    def get_bond_vecs(coords3d):
        return coords3d[constraints[:, 0]] - coords3d[constraints[:, 1]]

    # Original bond lengths
    lengths = np.linalg.norm(get_bond_vecs(coords3d), axis=1)
    lengths_sq = lengths**2

    if energy_forces_getter is None:
        energy_forces_getter = energy_forces_getter_closure(geom)
        # def forces_getter(coords):
            # return geom.get_energy_and_forces_at(coords)["forces"]

    def rattle(coords, velocities, forces):
        acceleration = forces * inv_masses_rep * FORCE2ACC
        # Bond vectors at time t
        bond_vecs = get_bond_vecs(coords.reshape(-1, 3))

        # Unconstrained position update
        coords3d_updated = (coords + dt*velocities + dt2_sq * acceleration).reshape(-1, 3)
        # Unconstrained velocity update
        velocities3d_updated = (velocities + dt2 * acceleration).reshape(-1, 3)
        if remove_com_v:
            velocities3d_updated = remove_com_velocity(velocities3d_updated, masses)

        # First part of RATTLE algorithm.
        #   Yields updated positions for t+dt and half-updated velocities for t+dt/2
        for i in range(max_cycles):
            corrected = False
            for j, (atom_1, atom_2) in enumerate(constraints):
                bond_vec_updated = coords3d_updated[atom_1] - coords3d_updated[atom_2]
                bond_length_sq = bond_vec_updated.dot(bond_vec_updated)
                ref_length_sq = lengths_sq[j]
                diff_sq = ref_length_sq - bond_length_sq

                # We have to correct coordinates and velocities if the deviations is
                # above the tolerance.
                if abs(diff_sq) <= (ref_length_sq * tol_sq):
                    continue

                corrected = True
                # Otherwise try to satisfy the constraint by calculating the
                # approximate lagrange multiplier 'g'.
                dot = bond_vec_updated.dot(bond_vecs[j])
                # TODO: test for constraint failure
                g = diff_sq / (2 * inv_masses_sums[j] * dot)
                
                # Update positions to satify constraint
                atom_1_factor = g * bond_vecs[j] * inv_masses[atom_1]
                atom_2_factor = g * bond_vecs[j] * inv_masses[atom_2]
                coords3d_updated[atom_1] += atom_1_factor
                coords3d_updated[atom_2] -= atom_2_factor
                # Update velocities to satify constraint
                velocities3d_updated[atom_1] += atom_1_factor / dt
                velocities3d_updated[atom_2] -= atom_2_factor / dt

            # Stop macro-iterations when no correction was done
            if not corrected:
                # print(f"RATTLE_1 finished after {i+1} macro cycle(s)!")
                break
        else:
            raise Exception("First part of RATTLE did not converge!")

        # Calculate energy (E_pot) and forces at updated coordinates
        energy_new, forces_new = energy_forces_getter(coords3d_updated.flatten())
        velocities3d_updated += dt2 * (forces_new * inv_masses_rep * FORCE2ACC).reshape(-1, 3)
        if remove_com_v:
            velocities3d_updated = remove_com_velocity(velocities3d_updated, masses)
        # The coordinates aren't updated anymore in the second part, so we can
        # calculate the bond vectors here once and store them.
        bond_vecs = get_bond_vecs(coords3d_updated)

        # Second part of RATTLE algorithm
        #   Update velocities to full time step t+dt
        for i in range(max_cycles):
            corrected = False
            for j, (atom_1, atom_2) in enumerate(constraints):
                velocity_diff = (velocities3d_updated[atom_1]
                                 - velocities3d_updated[atom_2])
                dot = bond_vecs[j].dot(velocity_diff)
                # Approximate Lagrange multiplier
                k = dot / (inv_masses_sums[j] * lengths_sq[j])

                if abs(k) <= tol_sq:
                    continue

                corrected = True
                # Update velocities to satify constraint
                velocities3d_updated[atom_1] -= k * bond_vecs[j] * inv_masses[atom_1]
                velocities3d_updated[atom_2] += k * bond_vecs[j] * inv_masses[atom_2]

            # Stop macro-iterations when no correction was done
            if not corrected:
                # print(f"RATTLE_2 finished after {i+1} macro cycle(s)!")
                break
        else:
            raise Exception("Second part of RATTLE did not converge!")

        return coords3d_updated.flatten(), velocities3d_updated.flatten(), energy_new, forces_new

    return rattle



# [1] http://dx.doi.org/10.1063/1.4952956
#     Lee-Ping Wang, 2016

import numpy as np

from pysisyphus.intcoords.Primitive import Primitive


class Translation(Primitive):
    """See (II. Theory) in [1], Eq. (2)"""

    def __init__(self, *args, **kwargs):
        kwargs["calc_kwargs"] = ("cart_axis",)
        super().__init__(*args, **kwargs)

    @staticmethod
    def _weight(atoms, coords3d, indices, f_damping):
        return 1

    @staticmethod
    def _calculate(coords3d, indices, gradient=False, cart_axis=0):
        value = coords3d[indices, cart_axis].mean()
        if gradient:
            row = np.zeros_like(coords3d)
            row[indices, cart_axis] = 1 / len(indices)
            row = row.flatten()
            return value, row
        return value

    @staticmethod
    def _jacobian(coords3d, indices, cart_axis):
        size = len(indices) * 3
        return np.zeros(size * size)


class TranslationX(Translation):
    cart_axis = 0


class TranslationY(Translation):
    cart_axis = 1


class TranslationZ(Translation):
    cart_axis = 2



#!/usr/bin/env python3
"""Smoke-test trajectory dump behavior for opt/tsopt commands."""

from __future__ import annotations

import os
import subprocess
import sys
import tempfile
import time
from dataclasses import dataclass
from pathlib import Path


REPO_ROOT = Path(__file__).resolve().parents[1]
CLI_MODULE = "pdb2reaction"
TIMEOUT_ENV = "PDB2REACTION_DUMP_CASE_TIMEOUT_SEC"
GENERIC_TIMEOUT_ENV = "DOCS_DUMP_CASE_TIMEOUT_SEC"
DEFAULT_CASE_TIMEOUT_SEC = 300.0


@dataclass(frozen=True)
class Case:
    name: str
    args: list[str]
    expect_present: tuple[str, ...]
    expect_absent: tuple[str, ...] = ()


def _write_tiny_xyz(path: Path) -> None:
    path.write_text(
        "\n".join(
            [
                "3",
                "water guess",
                "O 0.000000 0.000000 0.000000",
                "H 0.957200 0.000000 0.000000",
                "H -0.239987 0.927297 0.000000",
                "",
            ]
        ),
        encoding="utf-8",
    )


def _frame_count(xyz_path: Path) -> int:
    lines = xyz_path.read_text(encoding="utf-8").splitlines()
    i = 0
    n = 0
    while i < len(lines):
        if not lines[i].strip():
            i += 1
            continue
        nat = int(lines[i].strip())
        i += nat + 2
        n += 1
    return n


def _run_cli(args: list[str], timeout_sec: float | None = None) -> None:
    cmd = [sys.executable, "-m", CLI_MODULE, *args]
    try:
        proc = subprocess.run(
            cmd,
            cwd=REPO_ROOT,
            text=True,
            capture_output=True,
            timeout=timeout_sec,
        )
    except subprocess.TimeoutExpired as exc:
        raise RuntimeError(
            f"Command timed out after {timeout_sec} sec: {' '.join(cmd)}"
        ) from exc
    if proc.returncode != 0:
        tail = (proc.stdout + "\n" + proc.stderr)[-3000:]
        raise RuntimeError(f"Command failed: {' '.join(cmd)}\n{tail}")


def _check_runnable() -> tuple[bool, str]:
    probe = subprocess.run(
        [sys.executable, "-m", CLI_MODULE, "opt", "--help"],
        cwd=REPO_ROOT,
        text=True,
        capture_output=True,
    )
    probe_out = f"{probe.stdout}\n{probe.stderr}"
    if "Command 'opt' is unavailable" in probe_out or "Missing dependency:" in probe_out:
        return False, "opt command is unavailable in this environment."
    if probe.returncode != 0:
        return False, f"probe command failed with exit code {probe.returncode}."
    return True, "ok"


def _validate_case(case: Case, timeout_sec: float | None = None) -> None:
    _run_cli(case.args, timeout_sec=timeout_sec)
    out_dir = Path(case.args[case.args.index("--out-dir") + 1]).resolve()
    for rel in case.expect_present:
        p = out_dir / rel
        if not p.exists():
            raise RuntimeError(f"[{case.name}] expected '{rel}' to exist.")
        if p.suffix == ".xyz":
            n_frames = _frame_count(p)
            if n_frames <= 0:
                raise RuntimeError(f"[{case.name}] '{rel}' has no frames.")
    for rel in case.expect_absent:
        p = out_dir / rel
        if p.exists():
            raise RuntimeError(f"[{case.name}] expected '{rel}' to be absent.")


def main() -> int:
    runnable, reason = _check_runnable()
    if not runnable:
        print(f"[dump-smoke] skipped: {reason}")
        return 0

    timeout_raw = ""
    timeout_env = TIMEOUT_ENV
    for env_name in (GENERIC_TIMEOUT_ENV, TIMEOUT_ENV):
        raw = os.environ.get(env_name, "").strip()
        if raw:
            timeout_raw = raw
            timeout_env = env_name
            break
    timeout_sec = float(timeout_raw) if timeout_raw else DEFAULT_CASE_TIMEOUT_SEC
    if timeout_sec <= 0:
        timeout_sec = None

    if timeout_sec is None:
        print("[dump-smoke] per-case timeout: disabled")
    else:
        print(f"[dump-smoke] per-case timeout: {timeout_sec:.1f}s (env: {timeout_env})")

    with tempfile.TemporaryDirectory(prefix="pdb2_dump_smoke_") as td:
        base = Path(td)
        tiny_xyz = base / "tiny.xyz"
        _write_tiny_xyz(tiny_xyz)

        cases = [
            Case(
                name="opt_grad_dump",
                args=[
                    "opt",
                    "-i",
                    str(tiny_xyz),
                    "-q",
                    "0",
                    "-m",
                    "1",
                    "--opt-mode",
                    "grad",
                    "--max-cycles",
                    "1",
                    "--dump",
                    "--out-dir",
                    str(base / "opt_grad_dump"),
                ],
                expect_present=("optimization_trj.xyz",),
            ),
            Case(
                name="opt_hess_dump",
                args=[
                    "opt",
                    "-i",
                    str(tiny_xyz),
                    "-q",
                    "0",
                    "-m",
                    "1",
                    "--opt-mode",
                    "hess",
                    "--max-cycles",
                    "1",
                    "--dump",
                    "--out-dir",
                    str(base / "opt_hess_dump"),
                ],
                expect_present=("optimization_trj.xyz",),
            ),
            Case(
                name="tsopt_grad_dump",
                args=[
                    "tsopt",
                    "-i",
                    str(tiny_xyz),
                    "-q",
                    "0",
                    "-m",
                    "1",
                    "--opt-mode",
                    "grad",
                    "--max-cycles",
                    "1",
                    "--dump",
                    "--out-dir",
                    str(base / "tsopt_grad_dump"),
                ],
                expect_present=("optimization_trj.xyz", "optimization_all_trj.xyz"),
            ),
            Case(
                name="tsopt_hess_dump",
                args=[
                    "tsopt",
                    "-i",
                    str(tiny_xyz),
                    "-q",
                    "0",
                    "-m",
                    "1",
                    "--opt-mode",
                    "hess",
                    "--max-cycles",
                    "1",
                    "--dump",
                    "--out-dir",
                    str(base / "tsopt_hess_dump"),
                ],
                expect_present=("optimization_trj.xyz",),
                expect_absent=("optimization_all_trj.xyz",),
            ),
            Case(
                name="tsopt_hess_nodump",
                args=[
                    "tsopt",
                    "-i",
                    str(tiny_xyz),
                    "-q",
                    "0",
                    "-m",
                    "1",
                    "--opt-mode",
                    "hess",
                    "--max-cycles",
                    "1",
                    "--no-dump",
                    "--out-dir",
                    str(base / "tsopt_hess_nodump"),
                ],
                expect_present=(),
                expect_absent=("optimization_trj.xyz", "optimization_all_trj.xyz"),
            ),
            Case(
                name="tsopt_hess_dump_legacy_bool",
                args=[
                    "tsopt",
                    "-i",
                    str(tiny_xyz),
                    "-q",
                    "0",
                    "-m",
                    "1",
                    "--opt-mode",
                    "hess",
                    "--max-cycles",
                    "1",
                    "--dump",
                    "True",
                    "--out-dir",
                    str(base / "tsopt_hess_dump_legacy_bool"),
                ],
                expect_present=("optimization_trj.xyz",),
            ),
        ]

        for idx, case in enumerate(cases, start=1):
            print(f"[dump-smoke] case {idx}/{len(cases)}: {case.name}")
            started = time.perf_counter()
            _validate_case(case, timeout_sec=timeout_sec)
            elapsed = time.perf_counter() - started
            print(f"[dump-smoke] case ok: {case.name} ({elapsed:.1f}s)")

    print(f"[dump-smoke] validated {len(cases)} cases.")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



# [1] https://doi.org/10.1063/1.4894472
#     Diabatization based on the dipole and quadrupole: The DQ method
#     Hoyer, Xu, Ma, Gagliardi, Truhlar, 2014
# [2] https://doi.org/10.1063/1.4948728
#     The DQ and DQΦ electronic structure diabatization methods:
#     Validation for general applications
#     Hoyer, Parker, Galgiardi, Truhlar, 2016
# [3] https://doi.org/10.1063/1.3042233
#     Constructing diabatic states from adiabatic states: Extending
#     generalized Mulliken–Hush to multiple charge centers with Boys localization
#     Subotnik, Yeganeh, Cave, Ratner, 2008

from dataclasses import dataclass
from typing import Optional

from jinja2 import Template
import numpy as np
from numpy.typing import NDArray

from pysisyphus.helpers_pure import highlight_text, to_subscript_num
from pysisyphus.wavefunction import logger
from pysisyphus.wavefunction.localization import (
    JacobiSweepResult,
    jacobi_sweeps,
    get_fb_ab_func,
    get_fb_cost_func,
)

DQTemplate = Template(
    """{{ name }}

Dipole moments
--------------
{{ dip_moms }}
{%- if quad_moms is not none%}
Trace of quadrupole tensor
--------------------------
{{ quad_moms }}
α = {{ "%.2f"|format(alpha) }}
{% endif %}
{%- if epots is not none %}
Electronic component of electrostatic potential
-----------------------------------------------
{{ epots}}
β = {{ "%.2f"|format(beta) }}
{% endif %}

"""
)


def dq_jacobi_sweeps(
    dip_moms: NDArray[float],
    quad_moms: Optional[NDArray[float]] = None,
    epots: Optional[NDArray[float]] = None,
    alpha: Optional[float] = 10.0,
    beta: Optional[float] = 1.0,
) -> JacobiSweepResult:
    """Rotation matrix from DQ-diabatization as outlined in [1], [2] and [3].

    When no quadrupole moments are given, the DQ-diabatization reduces to a simple
    Boys-diabatization, as outlined by Subotnik et al in [3]. In this case we just
    zeros for the quadrupole moments. As the overall size of the matrices is small,
    the additional FLOPs dont hurt and the code can be kept simpler.

    We only use the trace of the quadrupole moment matrix. There are three
    dipole moment components, but only one trace of the quadrupole moment
    matrix.

    TODO: In principle this function can easily be extended to support an arbitrary
    number of properties, each with its own scaling factor. When this is implemented,
    we could drop the separate Boys-localiatzion function in wavefunction.localization.

    Parameters
    ----------
    dip_moms
        Dipole moment matrix of the adiabatic states.
        Shape (3, nstates, nstates).
    quad_moms
        Optional matrix containing the trace of the primitive quadrupole
        moments. Shape (1, nstates, nstates). Optional.
    epots
        Electronic part of the electrostatic potential.
    alpha
        Scaling factor for quadrupole moment contribution.
    beta
        Scaling factor for electrostatic potential contribution.

    Returns
    -------
    JacobiSweepResult
        Diabatization result.
    """

    _, nstates, _ = dip_moms.shape
    U = np.eye(nstates)

    assert dip_moms.shape == (3, *U.shape)
    assert U.ndim == 2
    assert U.shape[0] == U.shape[1]
    expected_quad_mom_shape = (1, *dip_moms.shape[1:])

    prefixes = [
        char_
        for property_, char_ in ((dip_moms, "D"), (quad_moms, "Q"), (epots, "Φ"))
        if property_ is not None
    ]
    name = "".join(prefixes) + "-diabatization"
    if no_quad_moms := quad_moms is None:
        quad_moms = np.zeros(expected_quad_mom_shape)
    if no_epots := epots is None:
        epots = np.zeros(expected_quad_mom_shape)
    quad_moms = quad_moms.reshape(*expected_quad_mom_shape)
    epots = epots.reshape(*expected_quad_mom_shape)

    dip_ab_func = get_fb_ab_func(dip_moms)
    quad_ab_func = get_fb_ab_func(quad_moms)
    epot_ab_func = get_fb_ab_func(epots)

    def ab_func(j, k, U):
        dip_A, dip_B = dip_ab_func(j, k, U)
        quad_A, quad_B = quad_ab_func(j, k, U)
        epot_A, epot_B = epot_ab_func(j, k, U)
        A = dip_A + alpha * quad_A + beta * epot_A
        B = dip_B + alpha * quad_B + beta * epot_B
        return A, B

    dip_cost_func = get_fb_cost_func(dip_moms)
    quad_cost_func = get_fb_cost_func(quad_moms)
    epot_cost_func = get_fb_cost_func(epots)

    def cost_func(U):
        dip_P = dip_cost_func(U)
        quad_P = quad_cost_func(U)
        epot_P = epot_cost_func(U)
        P = dip_P + alpha * quad_P + beta * epot_P
        return P

    msg = DQTemplate.render(
        name=highlight_text(name),
        dip_moms=dip_moms,
        quad_moms=quad_moms if not no_quad_moms else None,
        alpha=alpha,
        epots=epots if not no_epots else None,
        beta=beta,
    )
    logger.info(msg)
    return jacobi_sweeps(U, cost_func, ab_func)


DiaResultTemplate = Template(
    """
All energies are given in {{ unit }}.

Adiabatic energy matrix V
-------------------------
{{ adia_mat }}

Rotation matrix U
-----------------
{{ U }}
det(U)={{ "%.4f"|format(det) }}

Diabatic energy matrix D = UᵀVU
-------------------------------
{{ dia_mat }}

Diabatic states Ξᵢ sorted by energy
--------------------------------
{%- for ind, dia_state, dia_en in dia_states_sorted %}
{{ ind }}: {{ dia_state }}, {{ "%.4f"|format(dia_en) }} {{ unit }}
{%- endfor %}

Composition of diabatic states Ξᵢ
---------------------------------
{%- for dia_comp in dia_compositions %}
{{ dia_comp }}
{%- endfor %}

Weights U²
----------
{{ U2 }}

Absolute diabatic couplings
---------------------------
{%- for key, coupling in couplings %}
|{{ key }}| = {{ "%.4f"|format(coupling) }} {{ unit }}
{%- endfor %}
"""
)


@dataclass
class DiabatizationResult:
    U: NDArray[float]
    adia_ens: NDArray[float]
    dia_ens: NDArray[float]
    adia_mat: NDArray[float]
    dia_mat: NDArray[float]
    is_converged: bool
    cur_cycle: int
    P: float
    dip_moms: Optional[NDArray] = None
    quad_moms: Optional[NDArray] = None
    epots: Optional[NDArray] = None

    @property
    def nstates(self):
        return len(self.adia_ens)

    def __post_init__(self):
        assert len(self.adia_ens) == len(self.dia_ens)
        quad_shape = (self.nstates, self.nstates)
        assert self.U.shape == self.adia_mat.shape == self.dia_mat.shape == quad_shape

    @property
    def couplings(self):
        nstates = self.nstates
        abs_dia_mat = np.abs(self.dia_mat)
        couplings = dict()
        for from_ in range(nstates):
            for to_ in range(from_ + 1, nstates):
                key = (from_, to_)
                couplings[key] = couplings[key[::-1]] = abs_dia_mat[from_, to_]
        return couplings

    def render_report(self, adia_labels=None, unit="eV"):
        U = self.U
        nstates = self.nstates
        U2 = U ** 2
        det = np.linalg.det(U)

        subscripts = [to_subscript_num(i) for i in range(nstates)]
        adia_wfs = [f"Φ{subscripts[i]}" for i in range(nstates)]
        dia_wfs = [f"Ξ{subscripts[i]}" for i in range(nstates)]

        # Diabatic states, sorted by energy
        dia_ens = self.dia_ens
        sorted_inds = np.argsort(dia_ens)
        dia_states_sorted = list()
        for i, sort_ind in enumerate(sorted_inds):
            dia_states_sorted.append((i, dia_wfs[sort_ind], dia_ens[sort_ind]))

        # Composition of diabatic states
        dia_compositions = list()
        if adia_labels:
            adia_labels = [f"({al})" for al in adia_labels]
        else:
            adia_labels = ["" for _ in range(nstates)]
        for i in range(nstates):
            coeffs = U[:, i]
            signs = ["+" if c > 0 else "-" for c in coeffs]
            sum_ = " ".join(
                [
                    f"{sign} {c:>6.4f}·{awf}{al}"
                    for sign, c, awf, al in zip(
                        signs, np.abs(coeffs), adia_wfs, adia_labels
                    )
                ]
            )
            dia_compositions.append(f"{dia_wfs[i]} = {sum_}")

        # Couplings
        couplings = self.couplings
        flat_couplings = list()
        for from_ in range(nstates):
            for to_ in range(from_ + 1, nstates):
                key = f"D{subscripts[from_]}{subscripts[to_]}"
                flat_couplings.append((key, couplings[(from_, to_)]))

        rendered = DiaResultTemplate.render(
            unit=unit,
            adia_mat=self.adia_mat,
            U=U,
            det=det,
            dia_mat=self.dia_mat,
            dia_states_sorted=dia_states_sorted,
            dia_compositions=dia_compositions,
            U2=U2,
            couplings=flat_couplings,
        )
        return rendered


def dq_diabatization(adia_ens, dip_moms, quad_moms=None, epots=None, **kwargs):
    jac_res = dq_jacobi_sweeps(dip_moms, quad_moms=quad_moms, epots=epots, **kwargs)
    U = jac_res.C
    adia_mat = np.diag(adia_ens)
    dia_mat = U.T @ adia_mat @ U
    dia_ens = np.diag(dia_mat)
    dia_result = DiabatizationResult(
        U=U,
        adia_ens=adia_ens,
        dia_ens=dia_ens,
        adia_mat=adia_mat,
        dia_mat=dia_mat,
        is_converged=jac_res.is_converged,
        cur_cycle=jac_res.cur_cycle,
        P=jac_res.P,
        dip_moms=dip_moms,
        quad_moms=quad_moms,
        epots=epots,
    )
    return dia_result



import numpy as np

from pysisyphus.config import BEND_MIN_DEG, LB_MIN_DEG
from pysisyphus.helpers_pure import log
from pysisyphus.intcoords.eval import eval_primitives
from pysisyphus.intcoords.exceptions import NeedNewInternalsException
from pysisyphus.intcoords.valid import bend_valid, dihedral_valid
from pysisyphus.intcoords.PrimTypes import Bends, Dihedrals, Rotations


def correct_dihedrals(new_dihedrals, old_dihedrals):
    """Dihedrals are periodic. Going from -179° to 179° is not a step of 358°,
    but a step of 2°. By considering the actual distance of the dihedrals from
    π the correct step can be calculated.

        dihedral step length  = abs(abs(new_dihedral) - π) + abs(abs(old_dihedral) - π)

    or put differently

        dihedral step length = abs(abs(new_dihedral - old_dihedral) - 2*π)

    The sign is left to be determined. Going from -179° to 179° (roughly π - -π = 2π)
    is a counter clockwise rotation and the dihedral has to decrease below -π. Going
    from 179° to -179° (roughly -π - π = -2π) is a clockwise rotation and the dihedral
    increases abvove π. So the correct sign corresponds to the negative sign of the
    original difference.

    original difference  2π -> dihedral must decrease -> sign = -1
    original difference -2π -> dihedral must increase -> sign = +1

    Overall, the old dihedral is modified by the actual step length with the correct
    sign."""
    new_dihedrals = np.atleast_1d(new_dihedrals)
    old_dihedrals = np.atleast_1d(old_dihedrals)
    dihedrals_step = new_dihedrals - old_dihedrals
    shifted_by_2pi = np.abs(np.abs(dihedrals_step) - 2 * np.pi) < np.pi / 2
    corrected_dihedrals = new_dihedrals.copy()
    corrected_dihedrals[shifted_by_2pi] -= (
        2 * np.pi * np.sign(dihedrals_step[shifted_by_2pi])
    )
    return corrected_dihedrals


def update_internals(
    new_coords3d,
    old_internals,
    primitives,
    dihedral_inds,
    rotation_inds,
    bend_inds,
    check_dihedrals=False,
    check_bends=False,
    bend_min_deg=BEND_MIN_DEG,
    bend_max_deg=LB_MIN_DEG,
    rotation_thresh=0.9,
    logger=None,
):
    prim_internals = eval_primitives(new_coords3d, primitives)
    new_internals = np.array([prim_int.val for prim_int in prim_internals])
    internal_diffs = new_internals - old_internals

    new_rotations = new_internals[rotation_inds]
    # Check for approaching singularity as discussed in the geomeTRIC paper. The
    # original code seems to check this only for linear molecules and instead
    # does some +2π/-2π magic, similar to how dihedrals differences are handled.
    if (np.abs(new_rotations / np.pi) >= rotation_thresh).any():
        raise NeedNewInternalsException(new_coords3d)

    dihedrals = [prim_internals[i] for i in dihedral_inds]

    dihedral_diffs = internal_diffs[dihedral_inds]

    # Find differences that are shifted by 2*pi
    shifted_by_2pi = np.abs(np.abs(dihedral_diffs) - 2 * np.pi) < np.pi / 2
    new_dihedrals = np.array([dihed.val for dihed in dihedrals])
    if any(shifted_by_2pi):
        new_dihedrals[shifted_by_2pi] -= (
            2 * np.pi * np.sign(dihedral_diffs[shifted_by_2pi])
        )
        # Update values
        for dihed, new_val in zip(dihedrals, new_dihedrals):
            dihed.val = new_val

    invalid_inds = list()
    # See if dihedrals became invalid (collinear atoms)
    if check_dihedrals:
        are_valid = [dihedral_valid(new_coords3d, prim.inds) for prim in dihedrals]
        try:
            first_dihedral = dihedral_inds[0]
        except IndexError:
            first_dihedral = 0
        invalid_inds = [
            i + first_dihedral for i, is_valid in enumerate(are_valid) if not is_valid
        ]

    if check_bends and len(bend_inds) > 0:
        bends = [prim_internals[i] for i in bend_inds]
        are_valid = [
            bend_valid(new_coords3d, prim.inds, bend_min_deg, bend_max_deg)
            for prim in bends
        ]
        first_bend = bend_inds[0]
        invalid_inds = [
            i + first_bend for i, is_valid in enumerate(are_valid) if not is_valid
        ]

    if len(invalid_inds) > 0:
        invalid_prims = [primitives[i] for i in invalid_inds]
        invalid_msg = ", ".join([str(tp) for tp in invalid_prims])
        log(logger, "Internal coordinate(s) became invalid! Need new internal coordinates!")
        log(logger, f"Invalid primitives: {invalid_msg}")
        raise NeedNewInternalsException(
            new_coords3d, invalid_inds=invalid_inds, invalid_prims=invalid_prims
        )

    return prim_internals


def inds_from_prim_types(typed_prims, prim_types):
    inds = [
        i for i, (prim_type, *inds) in enumerate(typed_prims) if prim_type in prim_types
    ]
    return inds


def transform_int_step(
    int_step,
    old_cart_coords,
    cur_internals,
    Bt_inv_prim,
    primitives,
    typed_prims=None,
    dihedral_inds=None,
    rotation_inds=None,
    bend_inds=None,
    check_dihedrals=False,
    check_bends=False,
    bend_min_deg=BEND_MIN_DEG,
    bend_max_deg=LB_MIN_DEG,
    freeze_atoms=None,
    constrained_inds=None,
    update_constraints=False,
    cart_rms_thresh=1e-6,
    Bt_inv_prim_getter=None,
    max_cycles=25,
    logger=None,
):
    """Transformation is done in primitive internals, so int_step must be given
    in primitive internals and not in DLC!"""

    # If an iterable of typed prims is given we can derive bend/dihedral/rotation
    # indices from them.
    if typed_prims is not None:
        if dihedral_inds is None:
            dihedral_inds = inds_from_prim_types(typed_prims, Dihedrals)

        if bend_inds is None:
            bend_inds = inds_from_prim_types(typed_prims, Bends)

        if rotation_inds is None:
            rotation_inds = inds_from_prim_types(typed_prims, Rotations)

    if freeze_atoms is None:
        freeze_atoms = list()

    if constrained_inds is None:
        constrained_inds = list()

    freeze_atoms = np.array(freeze_atoms, dtype=int)
    new_cart_coords = old_cart_coords.copy()
    remaining_int_step = int_step
    target_internals = cur_internals + int_step
    # When we want to update the constraints we use the target primitive internals,
    if update_constraints:
        constrained_vals = target_internals[constrained_inds]
    # otherwise we try to stay with the original constrained values.
    else:
        constrained_vals = cur_internals[constrained_inds]

    def backtransform(remaining_int_step, Bt_inv_prim):
        """Separate function so it can be a called after the main loop
        finished to re-enforce the constraints."""
        nonlocal new_cart_coords

        cart_step = Bt_inv_prim.T.dot(remaining_int_step)
        # Remove step from frozen atoms.
        cart_step.reshape(-1, 3)[freeze_atoms] = 0.0
        cart_rms = np.sqrt(np.mean(cart_step**2))
        # Update cartesian coordinates
        new_cart_coords += cart_step
        # Determine new internal coordinates
        new_prim_ints = update_internals(
            new_cart_coords.reshape(-1, 3),
            old_internals,
            primitives,
            dihedral_inds,
            rotation_inds,
            bend_inds,
            check_dihedrals=check_dihedrals,
            check_bends=check_bends,
            bend_min_deg=bend_min_deg,
            bend_max_deg=bend_max_deg,
            logger=logger,
        )
        new_internals = [prim.val for prim in new_prim_ints]
        return new_prim_ints, new_internals, cart_rms

    last_rms = 9999
    old_internals = cur_internals
    backtransform_failed = True
    for i in range(max_cycles):
        if Bt_inv_prim_getter is not None:
            Bt_inv_prim = Bt_inv_prim_getter(new_cart_coords)
            log(logger, f"Recalculated (B^T)^+ in microcycle {i}")

        new_prim_ints, new_internals, cart_rms = backtransform(
            remaining_int_step, Bt_inv_prim
        )
        remaining_int_step = target_internals - new_internals
        internal_rms = np.sqrt(np.mean(remaining_int_step**2))
        log(
            logger,
            f"Cycle {i}: rms(Δcart)={cart_rms:1.4e}, rms(Δint.) = {internal_rms:1.5e}",
        )

        # This assumes the first cart_rms won't be > 9999 ;)
        if cart_rms < last_rms:
            # Store results of the conversion cycle for laster use, if
            # the internal-cartesian-transformation goes bad.
            best_cycle = (new_cart_coords.copy(), new_internals.copy())
            best_cycle_ind = i
        elif i != 0:
            # If the conversion somehow fails we fallback to the best previous step.
            log(logger, f"Backconversion failed! Falling back to step {best_cycle_ind}")
            new_cart_coords, new_internals = best_cycle
            break
        else:
            raise Exception(
                "Internal-cartesian back-transformation already "
                "failed in the first step. Aborting!"
            )
        old_internals = new_internals

        last_rms = cart_rms
        if cart_rms < cart_rms_thresh:
            log(
                logger,
                f"Internal->Cartesian transformation converged in {i+1} cycle(s)!",
            )
            backtransform_failed = False
            break

    if len(constrained_inds) > 0:
        for j in range(max_cycles):
            cur_constrained_vals = np.array(new_internals)[constrained_inds]
            diff = constrained_vals - cur_constrained_vals
            if any(np.abs(diff) <= 1e-5):
                break
            remaining_int_step = np.zeros_like(remaining_int_step)
            remaining_int_step[constrained_inds] = diff
            new_prim_ints, new_internals, _ = backtransform(
                remaining_int_step, Bt_inv_prim
            )
        if j > 0:
            log(logger, f"Re-enforced constraints in {j} additional cycle(s).")

    log(logger, "")

    # Return the difference between the new cartesian coordinates that yield
    # the desired internal coordinates and the old cartesian coordinates.
    cart_step = new_cart_coords - old_cart_coords
    return new_prim_ints, cart_step, backtransform_failed



from collections import namedtuple
import io
from pathlib import Path
import re
import shutil
import subprocess
import textwrap

import numpy as np
import pyparsing as pp

from pysisyphus.calculators.OverlapCalculator import OverlapCalculator
from pysisyphus.constants import AU2EV, BOHR2ANG
from pysisyphus.helpers_pure import file_or_str


class Gaussian16(OverlapCalculator):
    conf_key = "gaussian16"
    _set_plans = (
        {"key": "chk", "condition": lambda self: self.keep_chk},
        "fchk",
        ("fchk", "mwfn_wf"),
        "dump_635r",
    )

    def __init__(
        self,
        route,
        gbs="",
        gen="",
        keep_chk=False,
        stable="",
        fchk=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.route = route.lower()
        invalid_keywords = ("symmetry", "nosymm", "force", "opt", "freq", "irc")
        invalid_kws_given = [kw for kw in invalid_keywords if kw in self.route]
        assert (
            not invalid_kws_given
        ), f"Invalid keywords given in route: {invalid_kws_given}. Please remove them."
        self.gbs = gbs
        assert "@" not in gbs, "Give only the path to the .gbs file, " "without the @!"
        self.gen = gen
        self.keep_chk = keep_chk
        self.stable = stable
        self.fchk = fchk

        keywords = {
            kw: option
            for kw, option in [self.parse_keyword(kw) for kw in self.route.split()]
        }
        exc_keyword = [key for key in "td tda cis".split() if key in keywords]
        self.root = None
        self.nstates = None
        if exc_keyword:
            self.exc_key = exc_keyword[0]
            exc_dict = keywords[self.exc_key]
            self.nstates = int(exc_dict["nstates"])
            try:
                self.root = int(exc_dict["root"])
            except KeyError:
                self.root = 1
                self.log("No explicit root was specified! Using root=1 as default!")
            # Collect remaining options if specified
            self.exc_args = {
                k: v for k, v in exc_dict.items() if k not in ("nstates", "root")
            }
            # Delete exc keyword, as we build it later on
            self.route = re.sub(r"((?:td|cis|tda).+?(:?\s|$))", "", self.route)

        self.to_keep = ("com", "fchk", "log", "dump_635r", "input.xyz")
        if self.keep_chk:
            self.to_keep += (".chk",)

        self.fn_base = "gaussian16"
        self.inp_fn = f"{self.fn_base}.com"
        self.out_fn = f"{self.fn_base}.log"
        self.chk_fn = f"{self.fn_base}.chk"
        self.dump_base_fn = f"{self.fn_base}_rwfdump"

        self.gaussian_input = """
        %nproc={pal}
        %mem={mem}MB
        {chk_link0}
        {add_link0}
        #P {calc_type} {route} {exc}
        # Symmetry=None {reuse_data}
        
        title

        {charge} {mult}
        {coords}

        {gen}{gbs}



        """
        self.gaussian_input = textwrap.dedent(self.gaussian_input)

        self.parser_funcs = {
            "force": self.parse_force,
            "hessian": self.parse_hessian,
            "stable": self.parse_stable,
            "noparse": lambda path: None,
            "double_mol": self.parse_double_mol,
        }

        self.base_cmd = self.get_cmd("cmd")
        self.formchk_cmd = self.get_cmd("formchk")
        self.unfchk_cmd = self.get_cmd("unfchk")
        self.rwfdump_cmd = self.get_cmd("rwfdump")
        self.log(
            f"Using commands: g16={self.base_cmd}, formchk={self.formchk_cmd}, "
            f"unfchk={self.unfchk_cmd}, rwfdump={self.rwfdump_cmd}"
        )

    def make_exc_str(self):
        # Ground state calculation
        if not self.root:
            return ""
        root = f"root={self.root}"
        nstates = f"nstates={self.nstates}"
        pair2str = lambda k, v: f"{k}" + (f"={v}" if v else "")
        arg_str = ",".join([pair2str(k, v) for k, v in self.exc_args.items()])
        exc_str = f"{self.exc_key}=({root},{nstates},{arg_str})"
        return exc_str

    def reuse_data(self, path):
        # Nothing to reuse if no fchk or chk present
        if (self.fchk is None) and not hasattr(self, "chk"):
            return ""
        new_chk = path / self.chk_fn
        prev_fchk = new_chk.with_suffix(".fchk")
        shutil.copy(self.fchk, prev_fchk)
        cmd = f"{self.unfchk_cmd} {prev_fchk}".split()
        subprocess.run(cmd, stdout=subprocess.PIPE, cwd=path)
        self.log(f"Using MO guess from '{self.fchk}'.")

        reuse_str = "guess=read"
        # Also try to reuse information of previous TD calculation
        # if self.nstates and hasattr(self, "chk"):
        # DISABLED for now, as gaussian also resorts the states
        # internally, which then messes up our internal numbering
        # of states.
        # shutil.copy(self.chk, new_chk)
        # reuse_str += " td=read"
        # self.log("Using td=read")

        return reuse_str

    def make_gbs_str(self):
        if self.gbs:
            return f"@{self.gbs}"
        else:
            return ""

    def prepare_input(
        self, atoms, coords, calc_type, did_stable=False, point_charges=None
    ):
        path = self.prepare_path(use_in_run=True)
        xyz_str = self.prepare_xyz_string(atoms, coords)
        with open(path / "input.xyz", "w") as handle:
            handle.write(xyz_str)

        coords = self.prepare_coords(atoms, coords)
        kwargs = {
            "pal": self.pal,
            "mem": self.pal * self.mem,
            "chk_link0": f"%chk={self.chk_fn}",
            "add_link0": "",
            "route": self.route,
            # td/tda/cis(...)
            "exc": self.make_exc_str(),
            # guess=read td=read
            "reuse_data": self.reuse_data(path),
            "calc_type": calc_type,
            "charge": self.charge,
            "mult": self.mult,
            "coords": coords,
            "gbs": self.make_gbs_str(),
            "gen": self.gen,
        }
        if calc_type == "double_mol":
            update = {
                "chk_link0": "",
                "add_link0": "%KJob L302 1",
                "route": self.route + " iop(3/33=1) geom=notest",
                "calc_type": "",
                "reuse_data": "",
                "exc": "",
            }
            kwargs.update(update)
        # stable and td/tda can't be used together, so we omit the
        # td/tda string here
        if "stable" in calc_type.lower():
            kwargs["exc"] = ""
        # after a stability analysis we already got a converged wavefuntion
        # and we dont want to 'optimize' it any further
        if did_stable:
            kwargs["route"] += " scf=maxcycles=0"
        if point_charges is not None:
            kwargs["route"] += " charge"
            point_charges = point_charges.copy()
            point_charges[:, :3] *= BOHR2ANG
            with io.StringIO() as handle:
                np.savetxt(handle, point_charges, fmt="%16.10f")
                pc_str = handle.getvalue()
            # Append point charges to coords
            kwargs["coords"] += "\n\n" + pc_str
        inp = self.gaussian_input.format(**kwargs)
        return inp

    def make_fchk(self, path):
        cmd = f"{self.formchk_cmd} {self.chk_fn}".split()
        subprocess.run(cmd, stdout=subprocess.PIPE, cwd=path)
        self.log("Created .fchk")

    def run_rwfdump(self, path, rwf_index, chk_path=None):
        if chk_path is None:
            chk_path = path / self.chk_fn
        dump_fn = path / f"{self.dump_base_fn}_dump_{rwf_index}"
        cmd = f"rwfdump {chk_path} {dump_fn} {rwf_index}".split()
        subprocess.run(cmd)
        self.log(f"Dumped {rwf_index} from .chk.")
        return dump_fn

    def run_after(self, path):
        chk_path = path / self.chk_fn
        if not chk_path.exists():
            self.log("No .chk file found.")
            return
        # Create the .fchk file so we can keep it and parse it later on.
        self.make_fchk(path)
        if self.track:
            self.run_rwfdump(path, "635r")
            self.nmos, self.roots = self.parse_log(path / self.out_fn)

    def parse_keyword(self, text):
        word = pp.Word(pp.alphanums + "-" + "/")

        keyword = word.setResultsName("keyword")
        equals = pp.Literal("=")
        option = pp.Group(
            word
            + pp.Suppress(pp.Optional(equals))
            + pp.Optional(word, default="")
            + pp.Suppress(pp.Optional(","))
        )
        options = (
            pp.Suppress(pp.Optional(equals))
            + pp.Suppress(pp.Optional("("))
            + pp.OneOrMore(option)
            + pp.Suppress(pp.Optional(")"))
        ).setResultsName("options")

        parser = keyword + pp.Optional(options, default=[])

        result = parser.parseString(text)
        as_dict = result.asDict()
        kw = as_dict["keyword"]
        opt_dict = {key: value for key, value in as_dict["options"]}
        return kw, opt_dict

    @staticmethod
    def parse_fchk(fchk_path, keys):
        with open(fchk_path) as handle:
            text = handle.read()

        def to_float(s, loc, toks):
            return float(toks[0])

        # Matches -4.10693837E-16 and 1.60184209E-15
        float_ = pp.Word(pp.nums + "+-E.").setParseAction(to_float)
        # Start with Empty so we can progessively build the
        # parser for all keys.
        parser = pp.Empty()

        def parser_for_key(key):
            return pp.Group(
                pp.Suppress(pp.SkipTo(key))
                + key
                + pp.restOfLine
                + pp.ZeroOrMore(float_)
            )

        for key in keys:
            parser += parser_for_key(key)
        results = parser.parseString(text)
        results_dict = {}
        for key, res in zip(keys, results):
            # This handles scalar entries like
            # Total Energy  [...] R     -9.219940072302333E+01
            if len(res) == 2:
                results_dict[key] = float(res[-1].split()[-1])
            # This handles matrices like
            # Cartesian Gradient [...] R   N=           9 \
            # [Matrix entries]
            if len(res) > 2:
                results_dict[key] = np.array(res[2:])
        return results_dict

    def parse_all_energies(self, fchk=None):
        if fchk is None:
            fchk = self.fchk
        with open(fchk) as handle:
            text = handle.read()

        float_ = r"([\d\-\+Ee\.]+)"
        gs_re = re.compile(r"SCF Energy\s+R\s+" + float_)
        gs_mobj = gs_re.search(text)
        gs_energy = float(gs_mobj[1])
        # cis_re = re.compile(r"CIS Energy\s+R\s+" + float_)
        # cis_mobj = cis_re.search(text)
        # cis_energy = float(cis_mobj[1])
        etrans_re = re.compile(
            r"ETran state values\s+R\s+N=\s+(\d+)([\d\-\.Ee\+\s]+)", re.DOTALL
        )
        etrans_mobj = etrans_re.search(text)
        try:
            etrans = etrans_mobj[2].strip().split()
            etrans = np.array(etrans, dtype=float).reshape(-1, 16)
            exc_energies = etrans[:, 0]
            all_energies = np.zeros(exc_energies.size + 1)
            all_energies[0] = gs_energy
            all_energies[1:] = exc_energies
        except TypeError:
            all_energies = np.array((gs_energy,))
        return all_energies

    def store_and_track(self, results, func, atoms, coords, **prepare_kwargs):
        if self.track:
            self.store_overlap_data(atoms, coords)
            if self.track_root():
                # Redo the calculation with the updated root
                results = func(atoms, coords, **prepare_kwargs)
        results["all_energies"] = self.parse_all_energies()
        return results

    def get_energy(self, atoms, coords, **prepare_kwargs):
        results = self.get_forces(atoms, coords, **prepare_kwargs)
        results = self.store_and_track(
            results, self.get_energy, atoms, coords, **prepare_kwargs
        )
        return results

    def get_forces(self, atoms, coords, **prepare_kwargs):
        did_stable = False
        if self.stable:
            is_stable = self.run_stable(atoms, coords)
            self.log(f"Wavefunction is now stable: {is_stable}")
            did_stable = True
        inp = self.prepare_input(
            atoms, coords, "force", did_stable=did_stable, **prepare_kwargs
        )
        kwargs = {
            "calc": "force",
        }
        results = self.run(inp, **kwargs)
        results = self.store_and_track(
            results, self.get_forces, atoms, coords, **prepare_kwargs
        )
        return results

    def get_hessian(self, atoms, coords, **prepare_kwargs):
        inp = self.prepare_input(atoms, coords, "freq", **prepare_kwargs)
        kwargs = {
            "calc": "hessian",
        }
        results = self.run(inp, **kwargs)
        results = self.store_and_track(
            results, self.get_hessian, atoms, coords, **prepare_kwargs
        )
        return results

    def run_stable(self, atoms, coords, **prepare_kwargs):
        inp = self.prepare_input(atoms, coords, self.stable, **prepare_kwargs)
        self.log(f"Running stability analysis with {self.stable}")
        kwargs = {
            "calc": "stable",
        }
        results = self.run(inp, **kwargs)
        return results

    def parse_stable(self, path):
        log_path = path / self.out_fn
        with open(log_path) as handle:
            text = handle.read()
        instab_re = "wavefunction.*instability.*"
        instab_mobj = re.search(instab_re, text)
        if instab_mobj:
            self.log("Found instability!")
        stable_re = "wavefunction is stable"
        mobj = re.search(stable_re, text)
        is_stable = bool(mobj)
        return is_stable

    def run_calculation(self, atoms, coords, **prepare_kwargs):
        inp = self.prepare_input(atoms, coords, "", **prepare_kwargs)
        kwargs = {
            "calc": "noparse",
        }
        results = self.run(inp, **kwargs)
        if self.track:
            self.store_overlap_data(atoms, coords)
            self.track_root()
            self.log(
                "This track_root() call is a bit superfluous as the "
                "as the result is ignored :)"
            )
        return results

    def run_double_mol_calculation(self, atoms, coords1, coords2):
        self.log("Running double molecule calculation")
        double_atoms = atoms + atoms
        double_coords = np.hstack((coords1, coords2))
        inp = self.prepare_input(double_atoms, double_coords, "double_mol")
        kwargs = {
            "calc": "double_mol",
            "keep": False,
            "inc_counter": False,
            "run_after": False,
        }
        double_mol_ovlps = self.run(inp, **kwargs)
        return double_mol_ovlps

    def parse_tddft(self, path):
        with open(path / self.out_fn) as handle:
            text = handle.read()
        td_re = r"Excited State\s*\d+:\s*[\.\w\?-]+\s*([\d\.-]+?)\s*eV"
        matches = re.findall(td_re, text)
        assert len(matches) == self.nstates
        # Excitation energies in eV
        exc_energies = np.array(matches, dtype=np.float64)
        # Convert to Hartree
        exc_energies /= AU2EV
        return exc_energies

    @file_or_str(".log", method=True)
    def parse_log(self, text):
        def parse(text, regex, func):
            mobj = re.search(regex, text)
            return func(mobj[1])

        # Depending on wether we did the calculation with td=read or not
        # roots will be at a different value. Without reading the CI coeffs
        # from the checkpoint Gaussian will calculate four times as much roots
        # as requested in the first iterations of the calculation. This will
        # lead to a much higher number of expected number of CI-coefficients
        # when parsing the 635r dump later on.
        roots_re = r"Root\s+(\d+)"
        roots = np.array(re.findall(roots_re, text), dtype=int).max()

        # NBasis=    16 NAE=    12 NBE=    12 NFC=     6 NFV=     0
        basis_re = "NBasis=(.+)NAE=(.+)NBE=(.+)NFC=(.+)NFV=(.+)"
        basis_mobj = re.search(basis_re, text)
        basis_funcs, alpha, beta, _, _ = [int(n) for n in basis_mobj.groups()]
        a_occ = alpha
        b_occ = beta
        a_vir = basis_funcs - a_occ
        b_vir = basis_funcs - b_occ
        restricted = alpha == beta
        act_re = "NROrb=(.*)NOA=(.*)NOB=(.*)NVA=(.*)NVB=(.*)"
        act_mobj = re.search(act_re, text)
        _, a_act, b_act, _, _ = [int(n) for n in act_mobj.groups()]
        a_core = a_occ - a_act
        b_core = b_occ - b_act

        NMOs = namedtuple(
            "NMOs",
            "a_core, a_occ a_act a_vir " "b_core b_occ b_act b_vir " "restricted",
        )

        nmos = NMOs(
            a_core, a_occ, a_act, a_vir, b_core, b_occ, b_act, b_vir, restricted
        )
        self.log(str(nmos))
        return nmos, roots

    def parse_635r_dump(self, dump_path, roots, nmos):
        self.log(f"Parsing 635r dump '{dump_path}'")
        with open(dump_path) as handle:
            text = handle.read()
        regex = r"read left to right\):\s*(.+)"
        mobj = re.search(regex, text, re.DOTALL)
        arr_str = mobj[1].replace("D", "E")
        # Drop the first 12 items as they are always 0
        tmp = np.array(arr_str.split()[12:], dtype=np.float64)

        # the core electrons are frozen in TDDFT/TDA
        expected = (nmos.a_act * nmos.a_vir + nmos.b_act * nmos.b_vir) * roots * 2
        self.log(
            f"Expecting {expected} ci coefficients in {dump_path}. "
            f"There are {tmp.size} items (including eigenvalues)."
        )
        coeffs = tmp[:expected]
        # 1. dim: X+Y, X-Y -> 2
        # 2. dim: roots -> variable, usually higher than Nstates as gaussian
        #         calculates more roots in the first iteration of TDDFT
        # 3. dim: alpha, beta -> 2
        # The remainder depends on the number of alpha and beta electrons.
        # The vectors are given for active orbs. x virt. obs that is the
        # core orbitals are neglected.
        if nmos.restricted:
            XpY, XmY = coeffs.reshape(2, roots, 2, -1)
            X = 0.5 * (XpY + XmY)
            # Within a TDA calculation XpY and XmY are the same and
            # Y will only contain zeros.
            Y = XpY - X
        # In the unrestricted case we can't use 2 for the 3. dimension anymore
        # (same number of alpha and beta electrons). The sizes for the respective
        # spins would be (a_act*a_vir) and (b_act*b_vir).
        else:
            raise Exception("Unrestricted not supported yet!")
        # Shapes of X and Y
        # 1. dim: roots (more than initially requested) -> variable
        # 2. dim: alpha, beta                           -> 2
        # 3. dim: act. x virt.                          -> variable

        # xpy_fn = f"{dump_fn}_XpY"
        # np.save(xpy_fn, XpY)
        # xmy_fn = f"{dump_fn}_XmY"
        # np.save(xmy_fn, XmY)
        # x_fn = f"{dump_fn}_X"
        # np.save(x_fn, X)

        # As we only handle restricted calculations for now the alpha
        # and beta part are identical. So we only keep the alpha part.
        if self.nmos.restricted:
            # Drop beta part and restrict to the requested states
            X = X[: self.nstates, 0, :]
            Y = Y[: self.nstates, 0, :]

        X_full = np.zeros((self.nstates, nmos.a_occ, nmos.a_vir))
        X_full[:, nmos.a_core :] = X.reshape(-1, nmos.a_act, nmos.a_vir)
        Y_full = np.zeros((self.nstates, nmos.a_occ, nmos.a_vir))
        Y_full[:, nmos.a_core :] = Y.reshape(-1, nmos.a_act, nmos.a_vir)

        return X_full, Y_full

    def prepare_overlap_data(self, path):
        # Parse X eigenvector from 635r dump
        X, Y = self.parse_635r_dump(self.dump_635r, self.roots, self.nmos)

        # From http://gaussian.com/cis/, Examples tab, Normalization:
        #
        # 'For closed shell calculations, the sum of the squares of the
        # expansion coefficients is normalized to total 1/2 (as the beta
        # coefficients are not shown).'
        #
        # Right now we only deal with restricted calculatios, so alpha == beta
        # and we ignore beta. So we are lacking a factor of sqrt(2). Another
        # option would be to normalize all states to 1.
        # ci_coeffs *= 2**0.5

        # Parse mo coefficients from .fchk file and write a 'fake' turbomole
        # mos file.
        keys = (
            "SCF Energy",
            "Alpha Orbital Energies",
            "Alpha MO coefficients",
            "ETran state values",
        )
        scf_key, mo_energies_key, mo_key, exc_key = keys
        fchk_dict = self.parse_fchk(self.fchk, keys=keys)
        mo_energies = fchk_dict[mo_energies_key]
        C = fchk_dict[mo_key].T
        C = C.reshape(-1, mo_energies.size)

        gs_energy = fchk_dict[scf_key]
        exc_data = fchk_dict[exc_key].reshape(-1, 16)
        exc_energies = exc_data[:, 0]
        all_energies = np.zeros(len(exc_energies) + 1)
        all_energies[0] = gs_energy
        all_energies[1:] += exc_energies
        return C, X, Y, all_energies

    def parse_force(self, path):
        results = {}
        keys = ("SCF Energy", "Total Energy", "Cartesian Gradient")
        fchk_path = Path(path) / f"{self.fn_base}.fchk"
        fchk_dict = self.parse_fchk(fchk_path, keys)
        results["energy"] = fchk_dict["SCF Energy"]
        results["forces"] = -fchk_dict["Cartesian Gradient"]

        if self.nstates:
            # This sets the proper excited state energy in the
            # results dict and also stores all energies.
            exc_energies = self.parse_tddft(path)
            # G16 root input is 1 based, so we substract 1 to get
            # the right index here.
            root_exc_en = exc_energies[self.root - 1]
            gs_energy = fchk_dict["SCF Energy"]
            # Add excitation energy to ground state energy.
            results["energy"] += root_exc_en
            # Create a new array including the ground state energy
            # to save the energies of all states.
            all_ens = np.full(len(exc_energies) + 1, gs_energy)
            all_ens[1:] += exc_energies
            results["all_energies"] = all_ens

        return results

    def parse_hessian(self, path):
        keys = (
            "Total Energy",
            "Cartesian Gradient",
            "Cartesian Force Constants",
        )
        fchk_path = Path(path) / f"{self.fn_base}.fchk"
        fchk_dict = self.parse_fchk(fchk_path, keys=keys)
        en_key, grad_key, hess_key = keys
        grad = fchk_dict[grad_key]
        tril_hess = fchk_dict[hess_key]
        tril_inds = np.tril_indices(grad.size)
        full_hessian = np.zeros((grad.size, grad.size))
        full_hessian[tril_inds] = tril_hess
        triu_inds = np.triu_indices(grad.size, k=1)
        full_hessian[triu_inds] = full_hessian.T[triu_inds]
        results = {
            "energy": fchk_dict[en_key],
            "forces": -grad,
            "hessian": full_hessian,
        }
        return results

    def parse_double_mol(self, path, out_fn=None):
        def repl_double(s, loc, toks):
            return toks[0].replace("D", "E")

        if out_fn is None:
            out_fn = self.out_fn

        with open(path / out_fn) as handle:
            text = handle.read()
        # Number of basis functions in the double molecule
        nbas = int(re.search(r"NBasis =\s*(\d+)", text)[1])
        assert nbas % 2 == 0
        # Gaussian prints columns of a triangular matrix including
        # the diagonal
        backup_white = pp.ParserElement.DEFAULT_WHITE_CHARS
        pp.ParserElement.setDefaultWhitespaceChars(" \t")
        int_ = pp.Suppress(pp.Word(pp.nums))

        float_ = pp.Word(pp.nums + ".D+-").setParseAction(repl_double)
        nl = pp.Suppress(pp.Literal("\n"))
        header = pp.OneOrMore(int_) + nl
        line = int_ + pp.OneOrMore(float_) + nl

        block = pp.Group(header + pp.OneOrMore(~header + line))

        parser = pp.Suppress(
            pp.SkipTo("*** Overlap *** \n", include=True)
        ) + pp.OneOrMore(block).setResultsName("blocks")

        result = parser.parseFile(path / out_fn)
        pp.ParserElement.setDefaultWhitespaceChars(backup_white)

        # The full double molecule overlap matrix (square)
        full_mat = np.zeros((nbas, nbas))

        def get_block_inds(block_ind, nbas, cols=5):
            """Returns the indices as required to assign the matrix
            printed by Gaussian. Gaussian prints a lower triangle
            matrix in blocks of five columns each."""
            start_row = block_ind * cols
            start_col = start_row
            inds = list()
            for row in range(start_row, nbas):
                col = start_col
                while (col <= row) and (col < start_row + cols):
                    inds.append((row, col))
                    col += 1
            return inds

        for i, block in enumerate(result["blocks"]):
            rows, cols = zip(*get_block_inds(i, nbas))
            full_mat[rows, cols] = block.asList()

        nbas_single = nbas // 2
        """The whole matrix consists of four blocks:
            Original overlaps of molecule 1
                b1 = full_mat[:nbas_single, :nbas_single]
            Zero, as we only get the lower triangle
                b2 = full_mat[:nbas_single, nbas_single:]
            Double molecule overlaps between basis functions of
            molecule 1 and molecule 2
                b3 = full_mat[nbas_single:, :nbas_single]
            Original overlaps of molecule 2
                b4 = full_mat[nbas_single:, nbas_single:]
        """
        double_mol_ovlp = full_mat[nbas_single:, :nbas_single]
        return double_mol_ovlp

    def parse_charges(self, path=None):
        if path is None and self.fchk is not None:
            fchk_path = self.fchk
        elif path is not None and path.endswith(".fchk"):
            fchk_path = path
        elif path is not None:
            fchk_path = Path(path) / f"{self.fn_base}.fchk"

        keys = ("Mulliken Charges",)
        fchk_dict = self.parse_fchk(fchk_path, keys=keys)
        charges = np.array(fchk_dict["Mulliken Charges"])

        return charges

    def get_chkfiles(self):
        return {
            "fchk": self.fchk,
        }

    def set_chkfiles(self, chkfiles):
        try:
            fchk = chkfiles["fchk"]
            self.fchk = fchk
            self.log(f"Set chkfile '{fchk}' as fchk.")
        except KeyError:
            self.log("Found no fchk information in chkfiles!")

    def __str__(self):
        return "Gaussian16 calculator"



import textwrap


class TablePrinter:
    def __init__(
        self,
        header,
        col_fmts,
        width=12,
        sub_underline=True,
        shift_left=0,
        fmts_update=None,
        mark="*",
    ):
        self.header = header
        self.col_fmts = col_fmts
        self.width = width
        self.sub_underline = sub_underline
        self.shift_left = shift_left
        if fmts_update is None:
            fmts_update = {}
        self.fmts_update = fmts_update
        self.mark = mark

        w = str(self.width - 1)

        self.fmts = {
            "int": "{:>" + w + "d}",
            "float": "{: >" + w + ".6f}",
            "float_short": "{: >" + w + ".2f}",
            "str": "{:>" + w + "s}",
            "mark": "{:>1s}",
        }
        self.fmts.update(fmts_update)
        if self.sub_underline:
            self.header = [h.replace("_", " ") for h in self.header]

        self.header_str = " ".join([h.rjust(self.width) for h in self.header])
        mark_fmt = self.fmts["mark"]
        self.conv_str = " ".join([self.fmts[fmt] + mark_fmt for fmt in self.col_fmts])
        h0_len = len(self.header[0])
        self.offset = self.width - h0_len
        self.prefix = " " * (self.offset - self.shift_left)
        self.sep = (
            self.prefix
            + "-" * (len(self.header_str) - self.width + h0_len)
            + "-" * abs(self.shift_left)
        )

    def print_sep(self):
        print(self.sep)

    def print_header(self, with_sep=True):
        print(self.header_str)
        if with_sep:
            self.print_sep()

    def print_row(self, args, marks=None):
        if marks is None:
            marks = ["" for _ in args]
        marked_args = list()
        for arg, to_mark in zip(args, marks):
            marked_args.append(arg)
            marked_args.append(self.mark if to_mark else " ")
        print(self.conv_str.format(*marked_args))

    def print(self, *args, **kwargs):
        text = " ".join([str(a) for a in args])
        try:
            level = kwargs["level"]
        except KeyError:
            level = 0
        level_prefix = "    " * level
        print(textwrap.indent(text, self.prefix + level_prefix))



import numpy as np

from pysisyphus.line_searches.LineSearch import LineSearch, \
                                                LineSearchNotConverged


class HagerZhang(LineSearch):

    def __init__(self, *args, alpha_prev=None, f_prev=None, dphi0_prev=None,
                 quad_step=False, eps=1e-6, theta=0.5, gamma=0.5, rho=5,
                 psi_0=.01, psi_1=.1, psi_2=2., psi_low=0.1, psi_hi=10,
                 Delta=.7, omega=1e-3, max_bisects=10, **kwargs):

        kwargs["cond"] = "wolfe"
        super().__init__(*args, **kwargs)

        self.alpha_prev = alpha_prev
        self.f_prev = f_prev
        self.dphi0_prev = dphi0_prev
        self.quad_step = quad_step

        self.eps = eps
        self.theta = theta
        self.gamma = gamma
        self.rho = rho
        self.psi_0 = psi_0
        self.psi_1 = psi_1
        self.psi_2 = psi_2
        self.psi_low = psi_low
        self.psi_hi = psi_hi
        self.Delta = Delta
        self.omega = omega
        self.max_bisects = max_bisects

    def prepare_line_search(self):
        super().prepare_line_search()

        self.epsk = self.eps * abs(self.get_fg("f", 0.))

    def bisect(self, a, b):
        """Bisect interval [a, b]."""
        for i in range(self.max_bisects):
            # U3 a.
            d = (1 - self.theta)*a + self.theta*b
            dphi_d = self.get_phi_dphi("g", d)
            if dphi_d >= 0:
                return a, d

            phi_d = self.get_phi_dphi("f", d)
            # U3 b.
            # If (dphi_d > 0) we would already have returned above...
            if phi_d <= self.phi0 + self.epsk:
                a = d
            # U3 c.
            elif phi_d > self.phi0 + self.epsk:
                b = d
        raise Exception("Bisect failed!")

    def interval_update(self, a, b, c):
        """Narrows down the bracketing interval."""
        # U0
        if not (a < c < b):
            return a, b

        phi_c, dphi_c = self.get_phi_dphi("fg", c)
        # U1, sign of slope projection changed. We already passed the minimum.
        if dphi_c >= 0:
            return a, c
        # U2, we are moving towards the minimum.
        elif phi_c <= self.phi0 + self.epsk:
            return c, b

        # U3, phi_c increased above phi0, so we probably passed the minimum.
        return self.bisect(a, c)

    def secant(self, a, b):
        """Take secant step."""
        dphia = self.get_phi_dphi("g", a)
        dphib = self.get_phi_dphi("g", b)
        return (a*dphib - b*dphia) / (dphib - dphia)

    def double_secant(self, a, b):
        """Take secant² step."""
        c = self.secant(a, b)
        A, B = self.interval_update(a, b, c)
        cB_close = np.isclose(c, B)
        cA_close = np.isclose(c, A)

        if cB_close:
            c_dash = self.secant(b, B)
        elif cA_close:
            c_dash = self.secant(a, A)

        if cB_close or cA_close:
            a_dash, b_dash = self.interval_update(A, B, c_dash)
        else:
            a_dash, b_dash = A, B
        return a_dash, b_dash

    def bracket(self, c):
        """Generate initial interval [a, b] that satisfies the opposite
        slope condition (dphi(a) < 0, dphi(b) > 0).
        """
        cs = list()
        p0epsk = self.phi0 + self.epsk
        for j in range(10):
            cs.append(c)

            dphi_j = self.get_phi_dphi("g", c)

            if (dphi_j >= 0) and (j == 0):
                return 0, c

            phi_j = self.get_phi_dphi("f", c)
            if dphi_j >= 0:
                phi_inds = np.array([self.get_fg("f", c) for c in cs[:-1]]) <= p0epsk
                # See https://stackoverflow.com/a/8768734
                ci = len(phi_inds) - phi_inds[::-1].argmax() - 1
                return cs[ci], c
            elif phi_j > p0epsk:
                return self.bisect(0, c)

            c *= self.rho

    def norm_inf(self, arr):
        """Returns infinity norm of given array."""
        return np.linalg.norm(arr, np.inf)

    def initial(self):
        """Get an initial guess for alpha."""
        if (~np.isclose(self.x0, np.zeros_like(self.x0))).any():
            c = self.psi_0 * self.norm_inf(self.x0)/self.norm_inf(self.g0)
        elif not np.isclose(self.f0, 0):
            c = self.psi_0 * self.f0 / self.norm_inf(self.g0)**2
        else:
            c = 1
        return c

    def take_quad_step(self, alpha, g0_):
        """Try to get alpha for minimum step from quadratic interpolation."""
        fact = max(self.psi_low, g0_/(self.dphi0*self.psi_2))
        alpha_ = min(fact, self.psi_hi) * alpha
        phi_ = self.get_phi_dphi("f", alpha_)
        denom = 2*((phi_-self.phi0)/alpha_ - self.dphi0)
        f_temp = self.get_fg("f", alpha_)
        if denom > 0.:
            c = -self.dphi0*alpha_ / denom
            if f_temp > self.get_fg("f", 0):
                c = max(c, alpha_*1e-10)
        else:
            c = alpha
        return c

    def run_line_search(self):
        if self.alpha_init is None and self.alpha_prev:
            alpha_init = self.alpha_prev
        elif self.alpha_init is None and self.alpha_prev is None:
            alpha_init = self.initial()
        else:
            alpha_init = self.alpha_init

        if self.quad_step:
            g0_ = -2*abs(self.get_fg("f", 0)/alpha_init) if (self.dphi0_prev is None) \
                  else self.dphi0_prev
            alpha_init = self.take_quad_step(self.psi_2*alpha_init, g0_)
        # This may raise LineSearchConverged
        _ = self.get_phi_dphi("fg", alpha_init)

        # TODO: cubic interpolation for better alpha_init
        ak, bk = self.bracket(alpha_init)
        for k in range(self.max_cycles):
            if self.cond_func(ak):
                break
            # secant² step
            a, b = self.double_secant(ak, bk)
            if (b - a) > self.gamma*(bk - ak):
                # Bisection step
                c = (a + b)/2
                a, b = self.interval_update(a, b, c)
            ak, bk = a, b
        else:
            raise LineSearchNotConverged

        return ak



from typing import Optional

import numpy as np

from pysisyphus.cos.ChainOfStates import ChainOfStates
from pysisyphus.cos.stiffness import get_stiff_stress

# [1] http://aip.scitation.org/doi/pdf/10.1063/1.1323224
#     10.1063/1.1323224
# [2] http://onlinelibrary.wiley.com/doi/10.1002/jcc.20780/pdf
#     10.1002/jcc.20780
# [3] https://aip.scitation.org/doi/10.1063/1.2841941
#     Sheppard, 2008
# [4] https://aip.scitation.org/doi/pdf/10.1063/1.1636455
#     Trygubenko, 2004
# [5] Nudged Elastic Band Method for Finding Minimum Energy Paths of Transitions
#     Hannes Jónsson , Greg Mills , Karsten W. Jacobsen
# https://github.com/cstein/neb/blob/master/neb/neb.py


class NEB(ChainOfStates):
    def __init__(
        self,
        images,
        variable_springs=False,
        k_max=0.3,
        k_min=0.1,
        perp_spring_forces=None,
        bandwidth: Optional[float] = None,
        **kwargs,
    ):
        super(NEB, self).__init__(images, **kwargs)

        assert k_max >= k_min, "k_max must be bigger or equal to k_min!"
        self.variable_springs = variable_springs
        self.k_max = k_max
        self.k_min = k_min
        self.perp_spring_forces = perp_spring_forces
        self.bandwidth = bandwidth

        self.delta_k = self.k_max - self.k_min
        self.k = list()

    def update_springs(self):
        # Check if there are enough springs
        if len(self.k) != len(self.images) - 1:
            self.k = np.full(len(self.images) - 1, self.k_min)
        if self.variable_springs:
            self.set_variable_springs()

    def set_variable_springs(self):
        shifted_energies = self.energy - self.energy.min()
        energy_max = max(shifted_energies)
        energy_ref = 0.85 * energy_max
        for i in range(len(self.k)):
            # The ith spring connects images i-1 and i.
            e_i = i + 1
            ith_energy = max(shifted_energies[e_i], shifted_energies[e_i - 1])
            if ith_energy < energy_ref:
                self.k[i] = self.k_min
            else:
                self.k[i] = self.k_max - self.delta_k * (energy_max - ith_energy) / (
                    energy_max - energy_ref
                )
        self.log("updated springs: " + self.fmt_k())

    def fmt_k(self):
        return ", ".join([str(f"{k:.03f}") for k in self.k])

    @property
    def parallel_forces(self):
        indices = range(len(self.images))
        par_forces = [self.get_parallel_forces(i) for i in indices]
        return np.array(par_forces).flatten()

    def get_spring_forces(self, i):
        if i not in self.moving_indices:
            return self.zero_vec.copy()

        if (i == 0) or (i == len(self.images) - 1):
            # We can't use the last image index because there is one
            # spring less than there are images.
            spring_index = min(i, len(self.images) - 2)
            return self.k[spring_index] * self.get_tangent(i)

        prev_coords = self.images[i - 1].coords
        ith_coords = self.images[i].coords
        next_coords = self.images[i + 1].coords
        spring_forces = self.k[i] * (next_coords - ith_coords) - (
            ith_coords - prev_coords
        )
        return spring_forces

    def get_quenched_dneb_forces(self, i):
        """See [3], Sec. VI and [4] Sec. D."""
        if not self.perp_spring_forces or (i not in self.moving_indices):
            return self.zero_vec.copy()
        forces = self.images[i].forces
        tangent = self.get_tangent(i)
        perp_forces = forces - forces.dot(tangent) * tangent
        spring_forces = self.get_spring_forces(i)
        tangent = self.get_tangent(i)
        perp_spring_forces = spring_forces - spring_forces.dot(tangent) * tangent
        dneb_forces = (
            perp_spring_forces - perp_spring_forces.dot(perp_forces) * perp_forces
        )
        perp_norm = np.linalg.norm(perp_forces)
        perp_spring_norm = np.linalg.norm(perp_spring_forces)

        # Switching function to quench the dneb forces
        # Eq. (15) in [3]
        #
        # If norm(perp_force) >> norm(perp_spring_forces): dneb_factor ~ 1
        # If norm(perp_force) << norm(perp_spring_forces): dneb_factor ~ 0
        #
        # If the perpendicular spring force is much bigger than the
        # perpendicular force the DNEB forces is nearly fully quenched.
        dneb_factor = 2 / np.pi * np.arctan2(perp_norm**2, perp_spring_norm**2)
        dneb_forces_quenched = dneb_factor * dneb_forces

        # An alternative switchting function is given in [5], Eq. (10)
        # f(phi) = 1/2 * (1 + cos(pi*cos(theta)))
        # f -> 0 for a straight path (theta -> 0°)
        # f -> 1 for a perpendicular path (theta -> 90°)
        # cos(theta) = (R_(i+1) - R_i) * (R_i - R_(i-1)) / (norm of numerator)

        return dneb_forces_quenched

    def get_parallel_forces(self, i):
        if i not in self.moving_indices:
            return self.zero_vec.copy()

        if (i == 0) or (i == len(self.images) - 1):
            # We can't use the last image index because there is one
            # spring less than there are images.
            spring_index = min(i, len(self.images) - 2)
            return self.k[spring_index] * self.get_tangent(i)

        prev_coords = self.images[i - 1].coords
        ith_coords = self.images[i].coords
        next_coords = self.images[i + 1].coords
        return (
            self.k[i]
            * (
                np.linalg.norm(next_coords - ith_coords)
                - np.linalg.norm(ith_coords - prev_coords)
            )
            * self.get_tangent(i)
        )

    # See https://stackoverflow.com/a/15786149
    # This way we can reuse the parents setter.
    @ChainOfStates.forces.getter
    def forces(self):
        if self._forces is not None:
            return self._forces

        org_results = self.calculate_forces()
        self.update_springs()
        indices = range(len(self.images))
        total_forces = np.array(
            [
                self.get_parallel_forces(i)
                + self.get_perpendicular_forces(i)
                + self.get_quenched_dneb_forces(i)
                for i in indices
            ]
        )
        total_forces = self.set_climbing_forces(total_forces)
        if self.bandwidth is not None:
            # kappa = self.k
            # breakpoint()
            stiff_stress = get_stiff_stress(
                bandwidth=self.bandwidth,
                kappa=self.k,
                image_coords=self.image_coords,
                tangents=self.get_tangents(),
            )
            total_forces = total_forces + stiff_stress
        total_forces[self.org_forces_indices] = org_results["forces"][
            self.org_forces_indices
        ]
        if self.org_forces_indices:
            self.log(
                f"Returning unrpojected original forces for image(s): {self.org_forces_indices}."
            )
        self._forces = total_forces.flatten()

        return self._forces



import numpy as np
import pyparsing as pp

from pysisyphus.elem_data import ATOMIC_NUMBERS, nuc_charges_for_atoms
from pysisyphus.helpers_pure import file_or_str
from pysisyphus.wavefunction import get_l, Shell, AOMixShells, Wavefunction
from pysisyphus.wavefunction.helpers import BFType

AOMIX_EXTS = ".in"


@file_or_str(*AOMIX_EXTS)
def parse_aomix(text):
    real = pp.common.real
    sreal = pp.common.sci_real
    int_ = pp.common.integer
    header = pp.CaselessLiteral("[AOMix Format]")
    title = pp.CaselessKeyword("[Title]")
    energy = pp.Suppress(pp.CaselessLiteral("[SCF Energy / Hartree]")) + sreal
    xyz = pp.Group(sreal + sreal + sreal)
    atom_line = pp.Group(
        pp.Word(pp.alphas).set_results_name("atom")
        + int_.set_results_name("id")
        + int_.set_results_name("atom_num")
        + xyz.set_results_name("xyz")
    )
    atoms = (
        pp.Suppress(pp.CaselessLiteral("[Atoms]"))
        + pp.Optional(pp.CaselessLiteral("AU")).set_results_name("unit")
        + pp.OneOrMore(atom_line).set_results_name("atom_lines")
    )
    prim_gto = pp.Group(sreal + sreal)
    contr_gto = pp.Group(
        pp.Word(pp.alphas).set_results_name("l")
        + int_.set_results_name("contr_depth")
        + real
        + pp.OneOrMore(prim_gto).set_results_name("prim_gtos")
    )
    center_contr_gto = pp.Group(
        int_.set_results_name("center")
        + int_
        + pp.OneOrMore(contr_gto).set_results_name("contr_gtos")
    )
    gtos = pp.CaselessLiteral("[GTO]") + pp.OneOrMore(
        center_contr_gto
    ).set_results_name("center_gtos")
    ao_label = pp.Word(pp.alphanums)
    mo_line = pp.Suppress(int_ + int_ + pp.Word(pp.alphas) + int_ + ao_label) + sreal
    sym_label = int_ + pp.one_of(("a", "b"), caseless=True)
    mo = pp.Group(
        pp.CaselessLiteral("Sym=")
        + sym_label.set_results_name("sym_label")
        + pp.CaselessLiteral("Ene=")
        + sreal.set_results_name("ene")
        + pp.CaselessLiteral("Spin=")
        + pp.one_of(("Alpha", "Beta"), caseless=True).set_results_name("spin")
        + pp.CaselessLiteral("Occup=")
        + sreal.set_results_name("occup")
        + pp.Group(pp.OneOrMore(mo_line)).set_results_name("coeffs")
    )
    mos = pp.CaselessLiteral("[MO]") + pp.OneOrMore(mo).set_results_name("mos")

    parser = (
        header
        + pp.Optional(title).set_results_name("title")
        + pp.Optional(energy).set_results_name("energy")
        + atoms
        + gtos
        + mos
    )
    result = parser.parseString(text)
    as_dict = result.asDict()

    return as_dict


def shells_from_aomix_dict(aomix_dict):
    atom_lines = aomix_dict["atom_lines"]
    center_gtos = aomix_dict["center_gtos"]
    _shells = list()
    for atom_line, cgtos in zip(atom_lines, center_gtos):
        atomic_num = ATOMIC_NUMBERS[atom_line["atom"].lower()]
        center_ind = cgtos["center"]
        assert atom_line["id"] == center_ind
        center_ind -= 1  # Start from 0 internally
        center = np.array(atom_line["xyz"])
        for cgto in cgtos["contr_gtos"]:
            exps, coeffs = zip(*cgto["prim_gtos"])
            exps = np.array(exps)
            coeffs = np.array(coeffs)
            L = get_l(cgto["l"])
            shell = Shell(
                L=L,
                atomic_num=atomic_num,
                center=center,
                coeffs=coeffs,
                exps=exps,
                center_ind=center_ind,
            )
            _shells.append(shell)
    shells = AOMixShells(_shells)
    return shells


def wavefunction_from_aomix_dict(aomix_dict, **wf_kwargs):
    shells = shells_from_aomix_dict(aomix_dict)

    atoms = list()
    coords3d = list()
    for atom_line in aomix_dict["atom_lines"]:
        atoms.append(atom_line["atom"].lower())
        coords3d.append(atom_line["xyz"])
    atoms = tuple(atoms)
    coords = np.array(coords3d).flatten()

    occ_a = 0
    occ_b = 0
    unrestricted = False
    C_a = list()
    C_b = list()
    for mo in aomix_dict["mos"]:
        spin = mo["spin"].lower()
        occup = mo["occup"]
        coeffs = mo["coeffs"]
        if spin == "alpha":
            occ_a += occup
            C_a.append(coeffs)
        elif spin == "beta":
            occ_b += occup
            C_b.append(coeffs)
        else:
            raise Exception(f"Unknown {spin=}!")

        if not unrestricted:
            unrestricted = spin == "beta"
    occ_a = int(occ_a)
    occ_b = int(occ_b)
    if not unrestricted:
        C_b = C_a.copy()
        occ_a = occ_b = occ_a // 2
        mult = 1
    else:
        mult = (occ_a - occ_b) + 1

    nuc_charge = nuc_charges_for_atoms(atoms).sum()
    charge = nuc_charge - occ_a - occ_b
    C = np.array((np.transpose(C_a), np.transpose(C_b)))
    _wf_kwargs = {
        "atoms": atoms,
        "coords": coords,
        "charge": charge,
        "mult": mult,
        "unrestricted": unrestricted,
        "occ": (occ_a, occ_b),
        "C": C,
        "bf_type": BFType.CARTESIAN,
        "shells": shells,
        **wf_kwargs,
    }

    _wf_kwargs.update(wf_kwargs)
    return Wavefunction(**_wf_kwargs)


@file_or_str(*AOMIX_EXTS)
def shells_from_aomix(text):
    aomix_dict = parse_aomix(text)
    return shells_from_aomix_dict(aomix_dict)


@file_or_str(*AOMIX_EXTS)
def wavefunction_from_aomix(text, **wf_kwargs):
    from time import time
    start = time()
    aomix_dict = parse_aomix(text)
    dur = time() - start
    print("dur", dur)
    return wavefunction_from_aomix_dict(aomix_dict, **wf_kwargs)



# [1] https://doi.org/10.1016/j.ccr.2018.01.019
#     Quantitative wave function analysis for excited states
#     of transition metal complexes
#     Mai et al., 2018

import itertools as it
from functools import partial
from typing import Dict, Iterable, List, Optional, Sequence, Tuple

import numpy as np
from numpy.typing import NDArray

from pysisyphus.wavefunction.wavefunction import Wavefunction


def norm_ci_coeffs(
    Xa: NDArray[float],
    Ya: NDArray[float],
    Xb: Optional[NDArray[float]] = None,
    Yb: Optional[NDArray[float]] = None,
    restricted_norm: float = 0.5,
    unrestricted_norm: float = 1.0,
) -> Tuple[NDArray[float], NDArray[float]]:
    """Normalize transition density matrices.

    target_norm = (N*X)**2 - (N*Y)**2
                = N**2 * (X**2 - Y**2)
    N**2 = target_norm / (X**2 - Y**2)
    N    = sqrt(target_norm / (X**2 - Y**2))
    """
    nstates_a, occ_a, virt_a = Xa.shape
    unrestricted = (Xb is not None) and (Yb is not None)
    if unrestricted:
        nstates_b, occ_b, virt_b = Xb.shape
        assert nstates_a == nstates_b
        assert (occ_a + virt_a) == (occ_b + virt_b)
        nstates = nstates_a
        X = np.concatenate((Xa.reshape(nstates, -1), Xb.reshape(nstates, -1)), axis=1)
        Y = np.concatenate((Ya.reshape(nstates, -1), Yb.reshape(nstates, -1)), axis=1)
        target_norm = unrestricted_norm
    else:
        nstates = nstates_a
        X = Xa.reshape(nstates, -1)
        Y = Ya.reshape(nstates, -1)
        target_norm = restricted_norm
    X_norms, Y_norms = [np.linalg.norm(mat, axis=1) for mat in (X, Y)]
    ci_norms = X_norms**2 - Y_norms**2
    N = np.sqrt(target_norm / ci_norms)
    X *= N[:, None]
    Y *= N[:, None]
    entries_a = occ_a * virt_a
    Xa = X[:, :entries_a].reshape(nstates, occ_a, virt_a)
    Ya = Y[:, :entries_a].reshape(nstates, occ_a, virt_a)
    returns = [Xa, Ya]
    if unrestricted:
        Xb = X[:, entries_a:].reshape(nstates, occ_b, virt_b)
        Yb = Y[:, entries_a:].reshape(nstates, occ_b, virt_b)
        returns.extend([Xb, Yb])
    return returns


def get_state_to_state_transition_density(
    T_a: NDArray[float], T_b: NDArray[float]
) -> NDArray[float]:
    """State-to-state transition density.

    Parameters
    ----------
    T_a
        Transition density matrix for state A of shape (occ, virt).
    T_b
        Transition density matrix for state B of shape (occ, virt).

    Returns
    -------
    T_ab
        State-to-state transition density of shape (occ+virt, occ+virt).
    """
    assert T_a.shape == T_b.shape

    electron = T_a.T @ T_b
    hole = T_b @ T_a.T
    occ, virt = T_a.shape
    ov = occ + virt
    T_ab = np.zeros((ov, ov))
    T_ab[:occ, :occ] = -hole
    T_ab[occ:, occ:] = electron
    return T_ab


def ct_numbers(
    C: NDArray[float], occ: int, Ts: NDArray[float], nfrags: int, frag_ao_map: Dict
) -> NDArray[float]:
    frag_inds = list(range(nfrags))
    T_full = np.zeros_like(C)
    u, _, vh = np.linalg.svd(C)
    lowdin = u @ vh  # (PQ^T) in eq. (17) of [1]

    omegas = list()
    for T_mo in Ts:
        T = T_full.copy()
        T[:occ, occ:] = T_mo
        Torth = lowdin @ T @ lowdin.T  # eq. (17) in [1]
        Torth2 = Torth**2
        state_omegas = list()
        # Pairs of fragments
        for i, j in it.product(frag_inds, frag_inds):
            aos_i = frag_ao_map[i]
            aos_j = frag_ao_map[j]
            om_ij = Torth2[aos_i][:, aos_j].sum()  # eq. (16) in [1]
            state_omegas.append(om_ij)
        omegas.append(state_omegas)
    return np.array(omegas)


def ct_numbers_for_states(
    wf: Wavefunction,
    frags: List[List[int]],
    Ta: NDArray[float],
    Tb: Optional[NDArray[float]] = None,
):
    """
    Charge-Transfer number for a given wavefunction and fragments.

    Parameters
    ----------
    wf
        Wavefunction object.
    frags
        List of lists of integers, containing the atom indices of the respective
        fragments.
    Ta
        Alpha-spin transition density matrices of shape (nstates, occ_a, virt_a).
    Tb
        Beta-spin transition density matrices of shape (nstates, occ_a, virt_a).

    Returns
    -------
    omegas
        Array of shape (nstates, len(frags)**2), containing the charge-transfer
        numbers of every state.
    """
    shells = wf.shells
    frag_ao_map = shells.fragment_ao_map(frags)
    nfrags = len(frags)

    occ_a, occ_b = wf.occ
    Ca, Cb = wf.C

    ctn = partial(ct_numbers, nfrags=nfrags, frag_ao_map=frag_ao_map)
    om_a = ctn(Ca, occ_a, Ta)
    if Tb is None:
        om_b = om_a
    else:
        om_b = ctn(Cb, occ_b, Tb)
    omegas = om_a + om_b
    return omegas


def mo_inds_from_tden(
    T: NDArray[float], ci_thresh: float
) -> Tuple[NDArray[int], NDArray[int], NDArray[float]]:
    """Indices (from, to) for CI-coefficients above a threshold."""
    assert T.ndim == 2
    occ, _ = T.shape
    from_, to_ = np.where(np.abs(T) > ci_thresh)
    coeffs = T[from_, to_]
    to_ += occ
    return from_, to_, coeffs


def to_pairs(arr_a: Iterable, arr_b: Iterable) -> List[frozenset]:
    """Zip two iterables to a list of frozensets."""
    return [frozenset(sorted(p)) for p in zip(arr_a, arr_b)]


def to_same_basis(
    pairs1: Sequence[frozenset],
    vals1: Iterable,
    pairs2: Sequence[frozenset],
    vals2: Iterable,
) -> Tuple[List[int], List[int], NDArray, NDArray]:
    unique_pairs = tuple(set(pairs1 + pairs2))
    zero_vec = np.zeros(len(unique_pairs))
    pair_map = {p: i for i, p in enumerate(unique_pairs)}

    def sort(pairs, vals):
        vals_sorted = zero_vec.copy()
        for pair, v in zip(pairs, vals):
            vals_sorted[pair_map[pair]] = v
        return vals_sorted

    vals1_sorted = sort(pairs1, vals1)
    vals2_sorted = sort(pairs2, vals2)
    try:
        from_, to_ = zip(*unique_pairs)
    # Raised when no unique pairs are present
    except ValueError:
        from_, to_ = list(), list()
    return from_, to_, vals1_sorted, vals2_sorted


def ovlp(
    from_A: NDArray[int],
    to_A: NDArray[int],
    vecA: NDArray[float],
    from_B: NDArray[int],
    to_B: NDArray[int],
    vecB: NDArray[float],
    S_MO: NDArray[float],
) -> NDArray[float]:
    """Overlap between two transition orbital pairs.

    According to Eq. (9)"""
    numA = vecA.size
    numB = vecB.size
    S = np.zeros((numA, numB))

    for i, coeff_a in enumerate(vecA):
        for j, coeff_b in enumerate(vecB):
            s_occ = S_MO[from_A[i], from_B[j]]
            s_vir = S_MO[to_A[i], to_B[j]]
            # Eq. (9)
            S[i, j] = coeff_a / abs(coeff_a) * coeff_b / abs(coeff_b) * s_occ * s_vir
    return S


def r_diff(
    vecP_from: NDArray[float], vec_to: NDArray[float], t_to: NDArray[float]
) -> float:
    abs_vec_to = np.abs(vec_to)

    def diff(vec_to):
        """Eq. (14)."""
        return np.abs(np.abs(vecP_from + vec_to) * t_to).sum()

    r_plus = diff(abs_vec_to)
    r_minus = diff(-1 * abs_vec_to)
    r = min(r_plus, r_minus)
    return r


def parse_tden(
    T: NDArray[float], thresh: float
) -> Tuple[List[frozenset], NDArray[int], NDArray[int], NDArray[float]]:
    from_T, to_T, coeffs = mo_inds_from_tden(T, thresh)
    pairs = to_pairs(from_T, to_T)
    return pairs, from_T, to_T, coeffs


def rAB(
    TXA: NDArray[float],
    TYA: NDArray[float],
    TXB: NDArray[float],
    TYB: NDArray[float],
    S_MO: NDArray[float],
    exc_thresh: float = 0.05,
    deexc_thresh: float = 0.02,
) -> float:
    """Transition-orbital-pair overlaps.

    See 'Transition orbital projection approach for excited state tracking' in
    J. Chem. Phys. 156, 214104 (2022); https://doi.org/10.1063/5.0081207

    Parameters
    ----------
    TXA
        Excitation coefficient (X) vector for 1 state at geometry A.
    TYA
        Deexcitation coefficient (Y) vector for 1 state at geometry A.
    TXB
        Excitation coefficient (X) vector for 1 state at geometry B.
    TYB
        Deexcitation coefficient (Y) vector for 1 state at geometry B.
    S_MO
        Matrix containing overlaps between the MOs at geometry A and B.
    exc_thresh
        Coefficient threshold up to which CI-coefficients from TXA and TXA
        are still considered.
    deexc_thresh
        Coefficient threshold up to which CI-coefficients from TYA and TYA
        are still considered.
    """
    # A
    a_pairs_A, a_from_A, a_to_A, alphaA = parse_tden(TXA, exc_thresh)
    b_pairs_A, b_from_A, b_to_A, betaA = parse_tden(TYA, deexc_thresh)
    a_from_A, a_to_A, alphaA, betaA = to_same_basis(a_pairs_A, alphaA, b_pairs_A, betaA)
    b_from_A = a_from_A
    b_to_A = a_to_A
    # B
    a_pairs_B, a_from_B, a_to_B, alphaB = parse_tden(TXB, exc_thresh)
    b_pairs_B, b_from_B, b_to_B, betaB = parse_tden(TYB, deexc_thresh)
    a_from_B, a_to_B, alphaB, betaB = to_same_basis(a_pairs_B, alphaB, b_pairs_B, betaB)
    b_from_B = a_from_B
    b_to_B = a_to_B

    # Eqs. (7) and (8)
    XA = alphaA + betaA
    YA = alphaA - betaA
    XB = alphaB + betaB
    YB = alphaB - betaB

    # Eqs. (10) and (11)
    SX = ovlp(a_from_A, a_to_A, XA, a_from_B, a_to_B, XB, S_MO)
    XPA = (np.abs(XA)[None, :] @ SX).flatten()
    XPB = (SX @ np.abs(XB)[:, None]).flatten()

    SY = ovlp(b_from_A, b_to_A, YA, b_from_B, b_to_B, YB, S_MO)
    YPA = (np.abs(YA)[None, :] @ SY).flatten()
    YPB = (SY @ np.abs(YB)[:, None]).flatten()

    # Eq. (13) in the original paper includes a factor 2, probably because the algorithm
    # is only formulated for closed-shell cases. Here, we exclude this factor and expect
    # that closed-shell/open-shell transition density matrices are normalized differently,
    # e.g. as done in 'norm_ci_coeffs'
    tA = alphaA**2 - betaA**2
    tB = alphaB**2 - betaB**2

    rXAB = r_diff(XPA, XB, tB)  # r_X^A->B
    rYAB = r_diff(YPA, YB, tB)  # r_Y^A->B
    rXBA = r_diff(XPB, XA, tA)  # r_X^B->A
    rYBA = r_diff(YPB, YA, tA)  # r_Y^B->A
    # Eq. (15)
    r = 0.5 * (rXAB + rYAB + rXBA + rYBA)
    return r


def top_differences(XA, YA, XB, YB, S_MO):
    states_A = XA.shape[0]
    states_B = XB.shape[0]

    rs = list()
    for xai, yai in zip(XA, YA):
        for xbj, ybj in zip(XB, YB):
            r = rAB(xai, yai, xbj, ybj, S_MO)
            rs.append(r)
    rs = np.array(rs).reshape(states_A, states_B)
    return rs


def tden_overlaps(
    C_1: NDArray[float],
    ci_coeffs1: NDArray[float],
    C_2: NDArray[:float],
    ci_coeffs2: NDArray[float],
    S_AO: NDArray[float],
):
    """
    Transition density overlaps.

    Parameters
    ----------
    mo_coeffs1 : ndarray, shape (MOs, AOs)
        MO coefficient matrix. MOs are expected to be in columns.
    mo_coeffs2 : ndarray
        See mo_coeffs1.
    ci_coeffs1 : ndarray, shape(occ. MOs, MOs)
        CI-coefficient matrix.
    ci_coeffs2 : ndarray, shape(occ. MOs, MOs)
        See ci_coeffs1.
    S_AO : ndarray, shape(AOs1, AOs2)
        Double molcule AO overlaps.
    """
    # Total number of molecular orbitals for ci_coeffs1 and ci_coeffs2 (occ + virt)
    nmo1, nmo2 = [sum(ci_coeffs.shape[1:]) for ci_coeffs in (ci_coeffs1, ci_coeffs2)]
    assert S_AO.shape == (nmo1, nmo2)
    _, occ1, _ = ci_coeffs1.shape
    _, occ2, _ = ci_coeffs2.shape
    # MO overlaps, and the respective sub-matrices (occ x occ), (virt x virt)
    S_MO = C_1.T @ S_AO @ C_2
    S_MO_occ = S_MO[:occ1, :occ2]
    S_MO_vir = S_MO[occ1:, occ2:]

    # Thanks Philipp and Klaus!
    overlaps = np.zeros((ci_coeffs1.shape[0], ci_coeffs2.shape[0]))
    for i, state1 in enumerate(ci_coeffs1):
        precontr = S_MO_vir.T @ state1.T @ S_MO_occ
        for j, state2 in enumerate(ci_coeffs2):
            overlaps[i, j] = np.trace(precontr @ state2)

    """
    overlaps = np.einsum(
        "mil,ij,njk,kl->mn",
        ci_coeffs1,
        S_MO_occ,
        ci_coeffs2,
        S_MO_vir.T,
        optimize=["einsum_path", (0, 3), (1, 2), (0, 1)],
    )
    """
    return overlaps


###############################
# Natural Transition Orbitals #
###############################


"""
def nto_overlaps(ntos_1, ntos_2, ao_ovlp):
    states1 = len(ntos_1)
    states2 = len(ntos_2)
    ovlps = np.zeros((states1, states2))
    for i in range(states1):
        n_i = ntos_1[i]
        l_i = n_i.lambdas[:, None]
        ntos_i = l_i * n_i.ntos
        for j in range(states2):
            n_j = ntos_2[j]
            l_j = n_j.lambdas[:, None]
            ntos_j = l_j * n_j.ntos
            ovlp = np.sum(np.abs(ntos_i.dot(ao_ovlp).dot(ntos_j.T)))
            ovlps[i, j] = ovlp
    return ovlps

def nto_org_overlaps(ntos_1, ntos_2, ao_ovlp, nto_thresh=0.3):
    states_1 = len(ntos_1)
    states_2 = len(ntos_2)
    ovlps = np.zeros((states_1, states_2))

    for i in range(states_1):
        n_i = ntos_1[i]
        l_i = n_i.lambdas[:, None]
        ntos_i = n_i.ntos[(l_i >= nto_thresh).flatten()]
        l_i_big = l_i[l_i >= nto_thresh]
        for j in range(states_2):
            n_j = ntos_2[j]
            l_j = n_j.lambdas[:, None]
            ntos_j = n_j.ntos[(l_j >= nto_thresh).flatten()]
            ovlp = np.sum(
                l_i_big[:, None] * np.abs(ntos_i.dot(ao_ovlp).dot(ntos_j.T))
            )
            ovlps[i, j] = ovlp
    return ovlps
"""



import re

import numpy as np

from pysisyphus.calculators.Calculator import Calculator


class DFTD3(Calculator):
    
    conf_key = "dftd3"

    def __init__(self, geom, functional, bjdamping=False, **kwargs):
        super().__init__(**kwargs)
        self.atoms = geom.atoms
        self.functional = functional.lower()
        self.bjdamping = bjdamping

        self.parser_funcs = {
            "energy": self.parse_energy,
            "grad": self.parse_gradient,
        }

    def parse_energy(self, path):
        with open(path / self.out_fn) as handle:
            text = handle.read()
        mobj = re.search(r"Edisp /kcal,au:\s+([\d\-\.]+)\s+([\d\-\.]+)", text)
        results={
            "energy": float(mobj.group(2)),
        }
        return results

    def parse_gradient(self, path):
        grad = np.loadtxt(path / "dftd3_gradient")
        energy = self.parse_energy(path)["energy"]
        results={
            "energy": energy,
            "grad": grad.flatten(),
        }
        return results
  
    def calc(self, coords3d, gradient=False):
        inp = self.prepare_turbo_coords(self.atoms, coords3d.flatten())
        
        args = ["-func", self.functional]
        if self.bjdamping:
            args.append("-bj")
        if gradient:
            args.append("-grad")
        
        if gradient:
            calc_type = "grad"
        else:
            calc_type = "energy"

        results = self.run(inp, calc=calc_type, cmd="dftd3", add_args=args)
        
        if not gradient:
            return results["energy"]
        else:
            return results["energy"], results["grad"] 
    
    def __str__(self):
        return f"DFT-D3({self.name})"






import numpy as np

try:
    from simtk.openmm import app
    from simtk import openmm as mm
    from simtk import unit
except ModuleNotFoundError:
    print("OpenMM is not installed!")

from pysisyphus.constants import BOHR2ANG
from pysisyphus.calculators.Calculator import Calculator


class OpenMM(Calculator):
    def __init__(self, topology, params, **kwargs):
        super().__init__(**kwargs)

        self.topology = topology
        self.params = params
        self.system = self.topology.createSystem(
            self.params, nonbondedMethod=app.NoCutoff
        )

        dummy_int = mm.VerletIntegrator(0.001)
        self.simulation = app.Simulation(self.topology, self.system, dummy_int)
        self.context = self.simulation.context

        self.en2au = 1 / unit.hartree / unit.AVOGADRO_CONSTANT_NA
        self.frc2au = 1 / (unit.hartree / unit.bohr) / unit.AVOGADRO_CONSTANT_NA

    def get_charges(self):
        (nb_force,) = [
            force
            for force in self.system.getForces()
            if force.__class__.__name__ == "NonbondedForce"
        ]
        charges = list()
        for i in range(nb_force.getNumParticles()):
            charge, *_ = nb_force.getParticleParameters(i)
            charges.append(charge.value_in_unit(unit.elementary_charge))
        charges = np.array(charges)
        return charges

    def report_charges(self):
        """Total charge of the system.

        Adapted from modeller.py _addIons() in OpenMM.
        """

        charges = self.get_charges()
        total_charge = charges.sum()
        charges_str = np.array2string(charges, precision=4)
        self.log(f"OpenMM charges: {charges_str}")
        self.log(f"OpenMM system total charge: {total_charge:.6f} e")

    def get_dipole_moment(self, coords3d, reference=None, masses=None):
        def get_com(coords3d):
            total_mass = sum(masses)
            com = (1 / total_mass * coords3d * masses[:, None]).sum(axis=0)
            return com

        charges = self.get_charges()
        ref_dict = {
            None: lambda c3d: c3d,
            "centroid": lambda c3d: c3d.mean(axis=1)[None, :],
            "com": lambda c3d: c3d - get_com(c3d)[None, :],
        }
        coords3d = ref_dict[reference](coords3d)
        # Dipole moment vector in bohr * elementary_charge
        dip_moms = (coords3d * charges[:, None]).sum(axis=0)
        tot_dip_mom = np.linalg.norm(dip_moms)
        return dip_moms, tot_dip_mom

    def get_energy(self, atoms, coords, **prepare_kwargs):
        results = self.get_forces(atoms, coords, **prepare_kwargs)
        return {"energy": results["energy"]}

    def get_forces(self, atoms, coords, **prepare_kwargs):
        positions = coords.reshape(-1, 3) * BOHR2ANG / 10  # 2d array in nm
        self.context.setPositions(positions)
        state = self.context.getState(getEnergy=True, getForces=True)
        energy = state.getPotentialEnergy() * self.en2au
        forces = state.getForces(asNumpy=True)
        forces = (forces * self.frc2au).reshape(-1)

        energy = float(energy)
        forces = np.array(forces)
        return {"energy": energy, "forces": forces}



from pysisyphus.optimizers.BacktrackingOptimizer import BacktrackingOptimizer


class SteepestDescent(BacktrackingOptimizer):
    def __init__(self, geometry, alpha=0.1, **kwargs):
        super().__init__(geometry, alpha=alpha, **kwargs)

    def prepare_opt(self):
        self.log("no backtracking in cycle 0")

    def optimize(self):
        if self.is_cos and self.align:
            self.procrustes()

        self.forces.append(self.geometry.forces)
        self.energies.append(self.geometry.energy)

        if self.cur_cycle > 0:
            self.skip = self.backtrack(self.forces[-1], self.forces[-2])

        step = self.alpha * self.forces[-1]
        step = self.scale_by_max_step(step)
        return step



from collections import namedtuple

import numpy as np

from pysisyphus.constants import ANG2BOHR as ANG2BOHR
from pysisyphus.Geometry import Geometry


ZLine = namedtuple(
    "ZLine", "atom rind r aind a dind d", defaults=(None, None, None, None, None, None)
)


def geom_from_zmat(
    zmat,
    atoms=None,
    coords3d=None,
    geom=None,
    start_at=None,
    drop_dummy=True,
    **geom_kwargs
):
    """Adapted from https://github.com/robashaw/geomConvert by Robert Shaw."""

    if isinstance(zmat, str):
        zmat = zmat_from_str(zmat)

    zmat_atoms = [zline.atom for zline in zmat]
    # Extend supplied geometry by zmat
    if geom is not None:
        atoms = geom.atoms
        coords3d = geom.coords3d
        start_at = len(geom.atoms)

    if atoms is not None:
        atoms = list(atoms) + zmat_atoms
    else:
        atoms = zmat_atoms

    # Grow coordindate array and assign old coordinates
    if coords3d is not None:
        _coords3d = np.zeros((len(coords3d) + len(zmat), 3))
        _coords3d[: len(coords3d)] = coords3d
        coords3d = _coords3d
    else:
        coords3d = np.zeros((len(zmat), 3), dtype=float)

    if atoms or coords3d:
        assert len(coords3d) == len(atoms)

    if start_at is None:
        start_at = 0

    for i, zline in enumerate(zmat, start_at):
        assert all(
            [
                (ind is None) or (ind >= 0)
                for ind in (zline.rind, zline.aind, zline.dind)
            ]
        ), "Found invalid atom index. Atom indices start with 1, not 0!"

        r = zline.r
        # First atom is placed at the origin
        if i == 0:
            continue
        # Bond along x-axis
        elif i == 1:
            coords3d[i, 0] = r
        # Angle in xy-plane from polar coordinates
        elif i == 2:
            r"""
            M       P <- add
             \     /
              u   v
               \ /
                O
            """
            theta = np.deg2rad(zline.a)
            # Center
            O = coords3d[zline.rind]
            # Bond, pointing away from O to M
            u = coords3d[zline.aind] - O
            # Direction of u along x axis (left/right)
            sign = np.sign(u[0])
            # Polar coordinates
            x = r * np.cos(theta)
            y = r * np.sin(theta)
            # Translate from center with correct orientation
            coords3d[i] = O + (sign * x, sign * y, 0.0)
        # Dihedral in xyz-space from spherical coordinates
        else:
            theta, phi = np.deg2rad((zline.a, zline.d))

            sin_theta = np.sin(theta)
            cos_theta = np.cos(theta)
            sin_phi = np.sin(phi)
            cos_phi = np.cos(phi)

            x = r * cos_theta
            y = r * sin_theta * cos_phi
            z = r * sin_theta * sin_phi

            r"""
            M <- add      N
             \           /
              u         v
               \       /
                O--w--P
            """

            O = coords3d[zline.rind]
            P = coords3d[zline.aind]
            N = coords3d[zline.dind]

            # Local axis system
            v_ = P - N
            v = v_ / np.linalg.norm(v_)
            w_ = O - P
            w = w_ / np.linalg.norm(w_)
            a = np.cross(v, w)
            a /= np.linalg.norm(a)
            b = np.cross(a, w)
            b /= np.linalg.norm(b)
            coords3d[i] = O - w * x + b * y + a * z

    if drop_dummy:
        atoms_ = list()
        coords3d_ = list()
        for atom, xyz in zip(atoms, coords3d):
            if atom.lower() == "x":
                continue
            atoms_.append(atom)
            coords3d_.append(xyz)
        atoms = atoms_
        coords3d = np.array(coords3d_)

    geom = Geometry(atoms, coords3d, **geom_kwargs)
    return geom


def geom_from_zmat_str(text, coord_type="cart", coord_kwargs=None):
    zmat = zmat_from_str(text)
    return geom_from_zmat(zmat, coord_type=coord_type, coord_kwargs=coord_kwargs)


def zmat_from_str(text):
    def dec(str_):
        return int(str_) - 1

    def to_bohr(str_):
        return float(str_) * ANG2BOHR

    def convert(items):
        funcs = (str, dec, to_bohr, dec, float, dec, float)
        return [f(item) for f, item in zip(funcs, items)]

    zmat = list()
    for line in text.strip().split("\n"):
        line = line.strip()
        if line.startswith("#"):
            continue
        zmat.append(ZLine(*convert(line.split())))
    return zmat


def zmat_from_fn(fn):
    with open(fn) as handle:
        text = handle.read()
    return zmat_from_str(text)


def geom_from_zmat_fn(fn, **geom_kwargs):
    zmat = zmat_from_fn(fn)
    geom = geom_from_zmat(zmat, **geom_kwargs)
    return geom



import struct
import socket

import numpy as np

from pysisyphus.socket_helper import send_closure, recv_closure, get_fmts


def ipi_client(
    addr, atoms, energy_getter, forces_getter, hessian_getter=None, hdrlen=12
):
    atom_num = len(atoms)
    # Number of entries in a Caretsian forces/coords vector
    cartesians = 3 * atom_num
    # Formats needed for struct.pack, to cast variables to bytes.
    # Bytes needed, to store a forces/coords vector
    floats_bytes = 8 * cartesians
    # Virial is hardcoded to the zero vector.
    VIRIAL = struct.pack("d" * 9, *np.zeros(9))
    ZERO = struct.pack("i", 0)

    fmts = get_fmts(cartesians)

    # Unix socket is hardcoded right now, but may also be inet-socket
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    addr = str(addr)
    sock.connect(addr)

    send_msg = send_closure(sock, hdrlen, fmts)
    recv_msg = recv_closure(sock, hdrlen, fmts)

    counter = 0
    while True:
        try:
            # Lets start talking
            recv_msg(expect="STATUS")  # The server initiates with a STATUS
            send_msg("READY")

            status = recv_msg(expect="STATUS")
            if status == "NEEDPOS":
                send_msg("HAVEPOS")
                _ = recv_msg(4, fmt="int")[0]  # Recive atom num from IPI
                coords = np.array(
                    recv_msg(floats_bytes, "floats")
                )  # Receive current coords
                assert coords.size % 3 == 0  # Assert Cartesian coordinates
                send_msg(atom_num, "int")
                # Just send back the current coordinates or translate all atoms in +X
                coords.reshape(-1, 3)[:, 0] += 1
                send_msg(coords, "floats")
                continue

            # When the optimization converged EXIT will be returned .. not documented!
            if status == "EXIT":
                print("Exited!")
                break

            # It seems we have to send READY two times ... not documented!
            send_msg("READY")
            # The server then returns POSDATA.
            recv_msg(expect="POSDATA")
            # Receive cell vectors, inverse cell vectors and number of atoms ...
            # but we don't use them here, so we don't even try to convert them to something.
            sock.recv(72)  # cell
            sock.recv(72)  # icell
            ipi_atom_num = recv_msg(4, fmt="int")[0]
            assert ipi_atom_num == atom_num
            # ... and the current coordinates.
            coords = np.array(recv_msg(floats_bytes, "floats"))

            recv_msg(expect="STATUS")
            # Indicate, that calculation is possible
            send_msg("HAVEDATA")
            get_what = recv_msg()
            # Acutal QC calculations
            if get_what == "GETENERGY":
                energy = energy_getter(coords)
                print(f"Calculated energy: {energy:.6f}, counter={counter}")
                send_msg("ENERGYREADY")
                send_msg(energy, "float")
            if get_what == "GETFORCE":
                forces, energy = forces_getter(coords)
                print(f"Calculated energy & forces: {energy:.6f}, counter={counter}")
                send_msg("FORCEREADY")
                # Send everything to the server
                send_msg(energy, "float")
                send_msg(atom_num, "int")
                send_msg(forces, "floats")
                send_msg(VIRIAL, packed=True)
                # We don't want to send additional information, so just send 0.
                send_msg(ZERO, packed=True)
            elif get_what == "GETHESSIAN":
                hessian, energy = hessian_getter(coords)
                hessian_packed = struct.pack("d" * cartesians ** 2, *hessian.flatten())
                print(f"Calculated energy & Hessian: {energy:.6f}, counter={counter}")
                send_msg("HESSIANREADY")
                # Send everything to the server
                send_msg(energy, "float")
                send_msg(atom_num, "int")
                send_msg(hessian_packed, packed=True)

            counter += 1
        except Exception as err:
            raise err


def calc_ipi_client(addr, atoms, calc, queue=None, **kwargs):
    assert calc is not None, "Supplied calculator must not be None!"

    def energy_getter(coords):
        if queue is not None:
            queue.put(("coords", coords))
        results = calc.get_energy(atoms, coords)
        energy = results["energy"]
        return energy

    def forces_getter(coords):
        if queue is not None:
            queue.put(("coords", coords))
        results = calc.get_forces(atoms, coords)
        forces = results["forces"]
        energy = results["energy"]
        return forces, energy

    def hessian_getter(coords):
        if queue is not None:
            queue.put(("coords", coords))
        results = calc.get_hessian(atoms, coords)
        hessian = results["hessian"]
        energy = results["energy"]
        return hessian, energy

    ipi_client(addr, atoms, energy_getter, forces_getter, hessian_getter, **kwargs)



#!/usr/bin/env python3
"""Validate local markdown links under docs/."""

from __future__ import annotations

import re
from pathlib import Path


REPO_ROOT = Path(__file__).resolve().parents[1]
DOCS_ROOT = REPO_ROOT / "docs"
LINK_RE = re.compile(r"(?<!\!)\[[^\]]+\]\(([^)]+)\)")
EXTERNAL_PREFIXES = ("http://", "https://", "mailto:", "tel:")


def _normalize_target(raw: str) -> str:
    target = raw.strip()
    if target.startswith("<") and target.endswith(">"):
        target = target[1:-1].strip()
    if " " in target:
        target = target.split(" ", 1)[0]
    return target


def _iter_links(path: Path) -> list[tuple[int, str]]:
    links: list[tuple[int, str]] = []
    in_fence = False
    for lineno, line in enumerate(path.read_text(encoding="utf-8").splitlines(), start=1):
        stripped = line.strip()
        if stripped.startswith("```"):
            in_fence = not in_fence
            continue
        if in_fence:
            continue
        for match in LINK_RE.finditer(line):
            links.append((lineno, _normalize_target(match.group(1))))
    return links


def _iter_toctree_targets(path: Path) -> list[tuple[int, str]]:
    targets: list[tuple[int, str]] = []
    in_toctree = False
    for lineno, line in enumerate(path.read_text(encoding="utf-8").splitlines(), start=1):
        stripped = line.strip()
        if stripped.startswith("```"):
            marker = stripped[3:].strip().lower()
            if not in_toctree and marker.startswith("{toctree}"):
                in_toctree = True
                continue
            if in_toctree:
                in_toctree = False
            continue
        if not in_toctree:
            continue
        if not stripped or stripped.startswith(":"):
            continue
        targets.append((lineno, stripped.split()[0]))
    return targets


def _resolve_doc_target(path: Path, target: str) -> Path | None:
    if not target or target.startswith(EXTERNAL_PREFIXES) or target.startswith("#"):
        return None
    if "*" in target:
        return None
    base = (path.parent / target).resolve()
    if base.suffix:
        return base
    for candidate in (base.with_suffix(".md"), base.with_suffix(".rst"), base):
        if candidate.exists():
            return candidate
    return base.with_suffix(".md")


def _check_path(path: Path, errors: list[str]) -> None:
    for lineno, target in _iter_links(path):
        if not target or target.startswith(EXTERNAL_PREFIXES):
            continue
        if target.startswith("#"):
            continue
        target_path = target.split("#", 1)[0]
        if not target_path:
            continue
        resolved = (path.parent / target_path).resolve()
        if not resolved.exists():
            errors.append(
                f"{path.relative_to(REPO_ROOT)}:{lineno}: broken local link -> {target}"
            )

    for lineno, target in _iter_toctree_targets(path):
        resolved = _resolve_doc_target(path, target)
        if resolved is None:
            continue
        if not resolved.exists():
            errors.append(
                f"{path.relative_to(REPO_ROOT)}:{lineno}: broken toctree target -> {target}"
            )


def main() -> int:
    errors: list[str] = []
    pages = sorted(DOCS_ROOT.rglob("*.md"))
    for md in pages:
        _check_path(md, errors)

    if errors:
        print("[link-check] failed:")
        for e in errors:
            print(f"- {e}")
        return 1

    print(f"[link-check] validated local links in {len(pages)} pages.")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from dataclasses import dataclass
from math import ceil
import tempfile
from typing import Tuple

import jinja2
import numpy as np
from numpy.typing import NDArray
import pyparsing as pp

from pysisyphus.elem_data import ATOMIC_NUMBERS, INV_ATOMIC_NUMBERS
from pysisyphus.Geometry import Geometry
from pysisyphus.helpers_pure import file_or_str
from pysisyphus.wrapper.jmol import view_cdd_cube


def get_grid(coords3d, num=10, offset=3.0):
    minx, miny, minz = coords3d.min(axis=0) - offset
    maxx, maxy, maxz = coords3d.max(axis=0) + offset
    X, Y, Z = np.mgrid[
        minx : maxx : num * 1j,
        miny : maxy : num * 1j,
        minz : maxz : num * 1j,
    ]
    xyz = np.stack((X.flatten(), Y.flatten(), Z.flatten()), axis=1)
    spacing = np.array((maxx - minx, maxy - miny, maxz - minz)) / (num - 1)
    return xyz, spacing, (num, num, num)


def get_grid_with_spacing(coords3d, spacing=0.30, margin=3.0):
    minx, miny, minz = coords3d.min(axis=0) - margin
    maxx, maxy, maxz = coords3d.max(axis=0) + margin
    dx = maxx - minx
    dy = maxy - miny
    dz = maxz - minz

    nx = ceil(dx / spacing)
    ny = ceil(dy / spacing)
    nz = ceil(dz / spacing)
    X, Y, Z = np.mgrid[
        minx : maxx : nx * 1j,
        miny : maxy : ny * 1j,
        minz : maxz : nz * 1j,
    ]
    xyz = np.stack((X.flatten(), Y.flatten(), Z.flatten()), axis=1)
    act_spacing = np.array(
        ((maxx - minx) / (nx - 1), (maxy - miny) / (ny - 1), (maxz - minz) / (nz - 1))
    )
    return xyz, act_spacing, (nx, ny, nz)


@dataclass
class Cube:
    atoms: Tuple
    coords3d: NDArray
    origin: NDArray
    npoints: NDArray
    axes: NDArray
    vol_data: NDArray
    comment1: str = "Generated by pysisyphus"
    comment2: str = ""

    @staticmethod
    def from_file(fn):
        return parse_cube(fn)

    @staticmethod
    def from_wf_and_grid(wf, vol_data, origin, axes):
        assert vol_data.ndim == 3, "Volume data 'vol_data' must be 3d!"
        return Cube(
            atoms=wf.atoms,
            coords3d=wf.coords3d,
            origin=origin,
            npoints=vol_data.shape,
            axes=axes,
            vol_data=vol_data,
        )

    def to_str(self):
        return cube_to_str(
            self.atoms,
            self.coords3d,
            self.vol_data.reshape(*self.npoints),
            self.origin,
            self.axes,
        )

    def write(self, fn):
        with open(fn, "w") as handle:
            handle.write(self.to_str())

    def view_cdd(self, **kwargs):
        with tempfile.NamedTemporaryFile("w", suffix=".cube") as tmp_cube:
            self.write(tmp_cube.name)
            view_cdd_cube(tmp_cube.name, **kwargs)


@file_or_str(".cube", ".cub")
def parse_cube(text):
    int_ = pp.common.integer
    sci_real = pp.common.sci_real

    def get_line_word(*args):
        return pp.Word(*args).setWhitespaceChars("\n")

    comment = get_line_word(pp.printables + " \t")
    cart_vec = pp.Group(sci_real + sci_real + sci_real)
    axis = pp.Group(int_ + cart_vec)
    atom_line = pp.Group(
        int_.set_results_name("atomic_num")
        + sci_real.set_results_name("charge")
        + cart_vec.set_results_name("coords")
        + pp.LineEnd()
    )

    parser = (
        comment.set_results_name("comment1")
        + comment.set_results_name("comment2")
        + int_.set_results_name("atom_num")
        + cart_vec.set_results_name("origin")
        + axis.set_results_name("axis1")
        + axis.set_results_name("axis2")
        + axis.set_results_name("axis3")
        + pp.ZeroOrMore(atom_line).set_results_name("atom_lines")
        + pp.ZeroOrMore(sci_real).set_results_name("vol_data")
    )
    res = parser.parseString(text).as_dict()

    comment1 = " ".join(res["comment1"])
    comment2 = " ".join(res["comment2"])
    atom_num = res["atom_num"]
    atom_lines = res["atom_lines"]
    assert atom_num == len(atom_lines)

    atom_nums, coords3d = zip(*[(al["atomic_num"], al["coords"]) for al in atom_lines])
    atoms = [INV_ATOMIC_NUMBERS[an].capitalize() for an in atom_nums]
    coords3d = np.array(coords3d)
    origin = np.array(res["origin"])

    npoints, axes = zip(*[res[k] for k in ("axis1", "axis2", "axis3")])
    npoints = np.array(npoints, int)
    assert all(npoints > 0), "Only Bohr are supported!"
    axes = np.array(axes, float)
    vol_data = np.array(res["vol_data"]).reshape(*npoints)

    cube = Cube(
        atoms=atoms,
        coords3d=coords3d,
        origin=origin,
        npoints=npoints,
        axes=axes,
        vol_data=vol_data,
        comment1=comment1,
        comment2=comment2,
    )

    return cube


@file_or_str(".cube", ".cub")
def geom_from_cube(text, **kwargs):
    cube = parse_cube(text)
    geom = Geometry(cube.atoms, cube.coords3d, **kwargs)
    return geom


CUBE_TPL = """{{ comment1 }}
{{ comment2 }}
{{ ifmt(atom_nums|length) }} {{ fmt(origin_x) }} {{ fmt(origin_y) }} {{ fmt(origin_z) }}
{% for n_points, (x, y, z) in zip(axes_npoints, axes) %}
{{ ifmt(n_points) }} {{ fmt(x) }} {{ fmt(y) }} {{ fmt(z) }}
{% endfor %}
{% for atomic_num, (x, y, z) in zip(atom_nums, coords3d) %}
{{ ifmt(atomic_num) }} 0.000000 {{ fmt(x) }} {{ fmt(y) }} {{ fmt(z) }}
{% endfor %}
{{ grid_str }}
"""


def cube_to_str(atoms, coords3d, vol_data, origin, axes):
    assert vol_data.ndim == 3
    env = jinja2.Environment(trim_blocks=True, lstrip_blocks=True)
    env.globals.update(zip=zip)
    tpl = env.from_string(CUBE_TPL)

    org_x, org_y, org_z = origin
    axes_npoints = vol_data.shape
    atom_nums = [ATOMIC_NUMBERS[atom.lower()] for atom in atoms]
    nx, ny, nz = axes_npoints
    grid_str = ""
    for x in range(nx):
        for y in range(ny):
            for z in range(nz):
                d = "{: >12.6e} ".format(vol_data[x, y, z])
                grid_str += d
                if z % 6 == 5:
                    grid_str += "\n"
            grid_str += "\n"

    def fmt(num):
        return f"{num: >12.6f}"

    def ifmt(num):
        return f"{num: >5d}"

    rendered = tpl.render(
        fmt=fmt,
        ifmt=ifmt,
        origin_x=org_x,
        origin_y=org_y,
        origin_z=org_z,
        comment1="Generated by pysisyphus",
        comment2=f"Total of {nx*ny*nz} points",
        axes=axes,
        axes_npoints=axes_npoints,
        atom_nums=atom_nums,
        coords3d=coords3d,
        grid_str=grid_str,
    )
    return rendered



import bisect
import itertools as it
import logging
import time

import numpy as np
import rmsd
from scipy.spatial.distance import pdist

from pysisyphus.calculators.XTB import XTB
from pysisyphus.helpers import check_for_end_sign
from pysisyphus.helpers_pure import highlight_text
from pysisyphus.intcoords.setup import get_pair_covalent_radii
from pysisyphus.optimizers.RFOptimizer import RFOptimizer
from pysisyphus.stocastic.align import matched_rmsd
from pysisyphus.xyzloader import make_trj_str_from_geoms


class Pipeline:

    def __init__(self, geom, calc_getter=None, seed=None, max_cycles=5, cycle_size=15,
                 rmsd_thresh=.1, energy_thresh=1e-3, energy_range=.125,
                 compare_num=25, break_after=2,
                 calc_kwargs=None):
        self.initial_geom = geom
        self.calc_getter = calc_getter
        self.max_cycles = max_cycles
        self.cycle_size = cycle_size
        if seed is None:
            self.seed = int(time.time())
        else:
            self.seed = seed
        self.rmsd_thresh = rmsd_thresh
        self.energy_thresh = energy_thresh
        self.energy_range = energy_range
        self.compare_num = compare_num
        self.break_after = break_after
        self.calc_kwargs = {
                "charge": 0,
                "mult": 1,
        }
        if calc_kwargs is not None:
            self.calc_kwargs.update(calc_kwargs)

        np.random.seed(self.seed)
        self.logger = logging.getLogger("stocastic")

        self.is_analytical2d = len(geom.atoms) == 1
        self.log(f"Seed: {self.seed}")
        self.coords_size = self.initial_geom.coords.size

        self.calc_counter = 0
        self.cur_cycle = 0
        self.cur_micro_cycle = 0
        # Indicates if the next cycle is the last one
        self.break_in = self.break_after

        self.new_geoms = []
        self.new_energies = []

        self.initial_coords3d = self.initial_geom.coords3d
        self.atoms = self.initial_geom.atoms

    def __str__(self):
        return f"{self.__class__.__name__}(seed={self.seed})"

    def log(self, message):
        """Write a log message.

        Wraps the logger variable.

        Parameters
        ----------
        message : str
            Message to be logged.
        """

        self.logger.debug(f"{message}")

    def get_valid_index_set(self, to_intersect):
        return set(range(len(self.new_energies))) & set(to_intersect)

    def atoms_are_too_close(self, geom, factor=.7):
        """Determine if atoms are too close."""
        dist_mat = pdist(geom.coords3d)
        cov_rad_mat = get_pair_covalent_radii(geom.atoms)
        too_close = dist_mat < factor*cov_rad_mat
        return any(too_close)

    def geom_is_close_in_energy(self, geom):
        energy = geom.energy
        i = bisect.bisect_left(self.new_energies, energy)
        # Determine if there are neighbours that are close in energy
        # as we insert left/before the most similary energy the indices
        # of the (to be) neighbours in the current self.new_energies list
        # are i-1 and i.
        valid_inds = self.get_valid_index_set((i-1, i))
        diffs = [abs(self.new_energies[j] - energy) for j in valid_inds]
        return len(diffs) > 0 and min(diffs) < self.energy_thresh

    def geom_is_new(self, geom):
        """Determine if geometry is not already known."""
        if len(self.new_geoms) == 0:
            self.log("Found first geometry!")
            return True

        i = bisect.bisect_left(self.new_energies, geom.energy)
        to_intersect = range(i-self.compare_num, i+self.compare_num)
        valid_inds = np.array(list(self.get_valid_index_set(to_intersect)))
        new_energies = np.array(self.new_energies)[valid_inds]
        # Restrict geometries for RMSD comparison to an energy range
        # around the energy of the geometry to check.
        in_range = np.abs(new_energies - geom.energy) < self.energy_range
        valid_inds = valid_inds[in_range]
        # If this evalutes to True the energy of the current in geometry is
        # quite different and we add the geometry.
        if valid_inds.size == 0:
            # print("Energy of geometry is very different from the remaining "
                  # "ones. Adding geometry!")
            reason = "different energy."
            is_new = True
        # Otherwise check the RMSD values for the remaining geometries that
        # are close in energy.
        else:
            rmsds = [matched_rmsd(geom, self.new_geoms[i])[0] for i in valid_inds]
            rmsds = np.array(rmsds)
            is_new = rmsds.min() > self.rmsd_thresh
            reason = f"different RMSD (min(RMSD) = {rmsds.min():.3f})"

        if is_new:
            self.log(f"Found new geometry based on {reason}")
        return is_new

    def geom_is_valid(self, geom):
        """Filter out geometries that are None, or were the atoms are too close
        or when they are already known."""

        valid = (geom is not None
                 and not self.geom_is_close_in_energy(geom)
        )

        if not self.is_analytical2d:
            valid = (valid
                     and not self.atoms_are_too_close(geom)
                     and self.geom_is_new(geom)
            )
        return valid
    
    def get_input_geom(self, geom):
        raise Exception("Implement me!")

    """
    def run_new_geom(self, geom):
        input_geom = self.get_input_geom(geom)
        input_coords = input_geom.coords.copy()

        # Check if the geometry is similar to an already known starting
        # geometry.
        overlaps = new_coords.dot(np.array(self.starting_coords).T)/self.coords_size
        # print("overlaps with already known starting coordinates")
        # print(overlaps)
        max_overlap_ind = overlaps.argmax()
        max_overlap = overlaps[max_overlap_ind]
        similar_fn = f"similar_{self.similar_ind:03d}_trj.xyz"
        # print(f"max overlap is {max_overlap:.1f}, {similar_fn}, index {max_overlap_ind}")
        max_coords = self.starting_coords[max_overlap_ind]
        self.similar_ind += 1
        rmsds = list()
        for sc in self.starting_coords:
            sc3d = sc.reshape(-1, 3)
            rm = rmsd.kabsch_rmsd(new_coords.reshape(-1,3), sc3d)
            rmsds.append(rm)
        rmsds = np.array(rmsds)
        self.starting_coords.append(new_coords)
    """

    def get_unique_geometries(self, geoms):
        geom_num = len(geoms)
        rmsds = np.full((geom_num, geom_num), np.inf)
        for i, j in it.combinations(range(geom_num), 2):
            coords1 = geoms[i].coords.reshape(-1, 3)
            coords2 = geoms[j].coords.reshape(-1, 3)
            rmsds[i, j] = rmsd.kabsch_rmsd(coords1, coords2)
        is_, js = np.where(rmsds < self.rmsd_thresh)
        similar_inds = np.unique(js)
        unique_geoms = [geoms[i] for i in range(geom_num) if i not in similar_inds]
        return unique_geoms

    def run_geom_opt(self, geom):
        if self.calc_getter is not None:
            calc = self.calc_getter(calc_number=self.calc_counter)
            geom.set_calculator(calc)
            opt_kwargs = {
                "gdiis": False,
                "thresh": "gau_loose",
                "overachieve_factor": 2,
                "max_cycles": 75,
            }
            opt = RFOptimizer(geom, **opt_kwargs)
            opt.run()
            opt_geom = geom if opt.is_converged else None
        else:
            calc = XTB(calc_number=self.calc_counter, **self.calc_kwargs)
            opt_result = calc.run_opt(geom.atoms, geom.coords, keep=False)
            try:
                opt_geom = opt_result.opt_geom
            except AttributeError:
                opt_geom = None

        self.calc_counter += 1
        return opt_geom

    def write_geoms_to_trj(self, geoms, fn, comments=None):
        with open(fn, "w") as handle:
            handle.write(
                make_trj_str_from_geoms(geoms, comments, energy_comments=True)
            )

    def run(self):
        for self.cur_cycle in range(self.max_cycles):
            cycle_start = time.time()
            self.log(highlight_text(f"Cycle {self.cur_cycle}"))
            input_geoms = [self.get_input_geom(self.initial_geom)
                           for _ in range(self.cycle_size)]
            # Write input geometries to disk
            self.write_geoms_to_trj(input_geoms, f"cycle_{self.cur_cycle:03d}_input_trj.xyz")
            # Run optimizations on input geometries
            calc_start = time.time()
            opt_geoms = list()
            for i, geom in enumerate(input_geoms, 1):
                print(f"Optimizing geometry {i:03d}/{self.cycle_size:03d}", end="\r")
                opt_geoms.append(self.run_geom_opt(geom))
            print()
            calc_end = time.time()
            calc_duration = calc_end - calc_start
            self.log(f"Optimizations took {calc_duration:.0f} s.")

            kept_geoms = list()
            for geom in opt_geoms:
                # Do all the filtering and reject all invalid geometries
                if not self.geom_is_valid(geom):
                    continue

                energy = geom.energy
                i = bisect.bisect_left(self.new_energies, energy)
                self.new_energies.insert(i, energy)
                self.new_geoms.insert(i, geom)
                kept_geoms.append(geom)
                if i == 0 and len(self.new_energies) > 1:
                    last_minimum = self.new_energies[1]
                    diff = abs(energy - last_minimum)
                    self.log(f"It is a new global minimum at {energy:.4f} au! "
                             f"Last one was {diff:.4f} au higher.")

            kept_num = len(kept_geoms)

            trj_filtered_fn = f"cycle_{self.cur_cycle:03d}_trj.xyz"
            # Sort by energy
            kept_geoms = sorted(kept_geoms, key=lambda g: g.energy)
            if kept_geoms:
                self.write_geoms_to_trj(kept_geoms, trj_filtered_fn)
                self.log(f"Kicks in cycle {self.cur_cycle} produced "
                         f"{kept_num} new geometries.")
                self.break_in = self.break_after
            elif self.break_in == 0:
                self.log("Didn't find any new geometries in the last "
                      f"{self.break_after} cycles. Exiting!")
                break
            else:
                self.log(f"Cycle {self.cur_cycle} produced no new geometries.")
                self.break_in -= 1

            cycle_end = time.time()
            cycle_duration = cycle_end - cycle_start
            self.log(f"Cycle {i} took {cycle_duration:.0f} s.")
            self.log("")
            if check_for_end_sign():
                break

        self.log(f"Run produced {len(self.new_energies)} geometries!")
        # Return empty list of nothing was found
        if not self.new_energies:
            return []

        fn = "final_trj.xyz"
        self.write_geoms_to_trj(self.new_geoms, fn)
        # self.new_energies = np.array(new_energies)
        np.savetxt("energies.dat", self.new_energies)
        first_geom = self.new_geoms[0]
        first_geom.standard_orientation()
        first_geom.energy = self.new_energies[0]

        if self.is_analytical2d:
            return self.new_geoms

        matched_geoms = [first_geom, ]
        for geom, energy in zip(self.new_geoms[1:], self.new_energies):
            rmsd, (_, matched_geom) = matched_rmsd(first_geom, geom)
            matched_geom.energy = energy
            matched_geoms.append(matched_geom)
        fn_matched = "final_matched_trj.xyz"
        self.write_geoms_to_trj(matched_geoms, fn_matched)
        return matched_geoms



# [1] Transition-State Optimization Methods using Internal Coordinates
#     https://macsphere.mcmaster.ca/handle/11375/15450?mode=full
#     Rabi, 2014, Phd Thesis

import numpy as np

from pysisyphus.intcoords.Primitive import Primitive
from pysisyphus.intcoords.derivatives import (
    q_rd1,
    dq_rd1,
    d2q_rd1,
    q_rd2,
    dq_rd2,
    d2q_rd2,
)


class RobustTorsion1(Primitive):
    @staticmethod
    def _weight(atoms, coords3d, indices, f_damping):
        return 1.0

    @staticmethod
    def _calculate(coords3d, indices, gradient=False):
        args = coords3d[indices].flatten()
        val = q_rd1(*args)

        if gradient:
            row = np.zeros_like(coords3d)
            grad = dq_rd1(*args).reshape(-1, 3)
            row[indices] = grad
            return val, row.flatten()
        return val

    @staticmethod
    def _jacobian(coords3d, indices):
        return d2q_rd1(*coords3d[indices].flatten())


class RobustTorsion2(Primitive):
    @staticmethod
    def _weight(atoms, coords3d, indices, f_damping):
        return 1.0

    @staticmethod
    def _calculate(coords3d, indices, gradient=False):
        args = coords3d[indices].flatten()
        val = q_rd2(*args)

        if gradient:
            row = np.zeros_like(coords3d)
            grad = dq_rd2(*args).reshape(-1, 3)
            row[indices] = grad
            return val, row.flatten()
        return val

    @staticmethod
    def _jacobian(coords3d, indices):
        return d2q_rd2(*coords3d[indices].flatten())



def red(str_):
    return f"\033[91m{str_}\033[0m"


def green(str_):
    return f"\033[92m{str_}\033[0m"


BOOL_COLORS = {
    True: green,
    False: red,
}


def bool_color(bool_, str_=None):
    if str_ is None:
        str_ = str(bool_)
    color = BOOL_COLORS[bool_]
    return color(str_)
